/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "quakedef.h"
#include "menu_common.h"
#include "winquake.h"


m_state_t m_state;

// EnterMenu wrappers for old functions; required for command support
// and for any other external calls
void M_Menu_Main_f (void) {menu_Main.EnterMenu ();}
void M_Menu_SinglePlayer_f (void) {menu_Singleplayer.EnterMenu ();}
void M_Menu_Load_f (void) {menu_Load.EnterMenu ();}
void M_Menu_Save_f (void) {menu_Save.EnterMenu ();}
void M_Menu_Options_f (void) {menu_Options.EnterMenu ();}
void M_Menu_Keys_f (void) {menu_Keybindings.EnterMenu ();}
void M_Menu_Video_f (void) {menu_Video.EnterMenu ();}
void M_Menu_Help_f (void) {menu_Help.EnterMenu ();}
void M_Menu_MultiPlayer_f (void) {menu_Multiplayer.EnterMenu ();}
void M_Menu_Setup_f (void) {menu_Setup.EnterMenu ();}
void M_Menu_ServerList_f (void) {}
void M_Menu_GameOptions_f (void) {}
void M_Menu_LanConfig666_f (void) {}
void M_Menu_Net_f (void) {}

m_state_t			m_return_state;

#define StartingGame	(m_multiplayer_cursor == 1)
#define JoiningGame		(m_multiplayer_cursor == 0)
#define SerialConfig	(m_net_cursor == 0)
#define DirectConfig	(m_net_cursor == 1)
#define	IPXConfig		(m_net_cursor == 2)
#define	TCPIPConfig		(m_net_cursor == 3)


/*
================
M_DrawCharacter

Draws one solid graphics character
================
*/
void M_DrawCharacter (int cx, int line, int num)
{
	Draw_Character (cx + ((vid.width - 320) >> 1), line, num);
}

void M_DrawBackwardsCharacter (int cx, int line, int num)
{
	Draw_BackwardsCharacter (cx + ((vid.width - 320) >> 1), line, num);
}

void M_Print (int cx, int cy, char *str)
{
	while (*str)
	{
		M_DrawCharacter (cx, cy, (*str)+128);
		str++;
		cx += 8;
	}
}

void M_PrintWhite (int cx, int cy, char *str)
{
	while (*str)
	{
		M_DrawCharacter (cx, cy, *str);
		str++;
		cx += 8;
	}
}

void M_PrintCenter (int cy, char *str)
{
	int cx = 160 - strlen (str) * 4;

	while (*str)
	{
		M_DrawCharacter (cx, cy, (*str)+128);
		str++;
		cx += 8;
	}
}

void M_PrintCenterWhite (int cy, char *str)
{
	int cx = 160 - strlen (str) * 4;

	while (*str)
	{
		M_DrawCharacter (cx, cy, *str);
		str++;
		cx += 8;
	}
}


void M_DrawPic (int x, int y, qpic_t *pic)
{
	Draw_Pic (x + ((vid.width - 320)>>1), y, pic);
}


//=============================================================================

/*
================
M_ToggleMenu_f
================
*/
void M_ToggleMenu_f (void)
{
	if (key_dest == key_menu)
	{
		if (m_state != m_main)
		{
			menu_Main.EnterMenu ();
			return;
		}

		key_dest = key_game;
		m_state = m_none;
		return;
	}

	if (key_dest == key_console)
		Con_ToggleConsole_f ();
	else menu_Main.EnterMenu ();
}



//=============================================================================
/* SEARCH MENU */

bool	searchComplete = false;
double		searchCompleteTime;

void M_Menu_Search_f (void)
{
	key_dest = key_menu;
	m_state = m_search;
	slistSilent = true;
	slistLocal = false;
	searchComplete = false;
	NET_Slist_f();

}


void M_Search_Draw (void)
{
	qpic_t	*p;
	int x;

	p = Draw_CachePic ("gfx/p_multi.lmp");
	M_DrawPic ( (320-p->width)/2, 4, p);
	x = (320/2) - ((12*8)/2) + 4;

	M_Print (x, 40, "Searching...");

	if(slistInProgress)
	{
		NET_Poll();
		return;
	}

	if (! searchComplete)
	{
		searchComplete = true;
		searchCompleteTime = realtime;
	}

	if (hostCacheCount)
	{
		M_Menu_ServerList_f ();
		return;
	}

	M_PrintWhite ((320/2) - ((22*8)/2), 64, "No Quake servers found");
	if ((realtime - searchCompleteTime) < 3.0)
		return;

	M_Menu_LanConfig666_f ();
}


void M_Search_Key (int key)
{
}

//=============================================================================
/* SLIST MENU */

int		slist_cursor;
bool slist_sorted;

void M_Menu_ServerList666_f (void)
{
	key_dest = key_menu;
	m_state = m_slist;
	slist_cursor = 0;
	slist_sorted = false;
}


void M_ServerList_Draw (void)
{
	int		n;
	char	string [64];
	qpic_t	*p;

	if (!slist_sorted)
	{
		if (hostCacheCount > 1)
		{
			int	i,j;
			hostcache_t temp;
			for (i = 0; i < hostCacheCount; i++)
				for (j = i+1; j < hostCacheCount; j++)
					if (strcmp(hostcache[j].name, hostcache[i].name) < 0)
					{
						Q_memcpy(&temp, &hostcache[j], sizeof(hostcache_t));
						Q_memcpy(&hostcache[j], &hostcache[i], sizeof(hostcache_t));
						Q_memcpy(&hostcache[i], &temp, sizeof(hostcache_t));
					}
		}
		slist_sorted = true;
	}

	p = Draw_CachePic ("gfx/p_multi.lmp");
	M_DrawPic ( (320-p->width)/2, 4, p);
	for (n = 0; n < hostCacheCount; n++)
	{
		if (hostcache[n].maxusers)
			sprintf(string, "%-15.15s %-15.15s %2u/%2u\n", hostcache[n].name, hostcache[n].map, hostcache[n].users, hostcache[n].maxusers);
		else
			sprintf(string, "%-15.15s %-15.15s\n", hostcache[n].name, hostcache[n].map);
		M_Print (16, 32 + 8*n, string);
	}
	M_DrawCharacter (0, 32 + slist_cursor*8, 12+((int)(realtime*4)&1));
}


void M_ServerList_Key (int k)
{
	switch (k)
	{
	case K_ESCAPE:
		M_Menu_LanConfig666_f ();
		break;

	case K_SPACE:
		M_Menu_Search_f ();
		break;

	case K_UPARROW:
	case K_LEFTARROW:
		S_LocalSound ("misc/menu1.wav");
		slist_cursor--;
		if (slist_cursor < 0)
			slist_cursor = hostCacheCount - 1;
		break;

	case K_DOWNARROW:
	case K_RIGHTARROW:
		S_LocalSound ("misc/menu1.wav");
		slist_cursor++;
		if (slist_cursor >= hostCacheCount)
			slist_cursor = 0;
		break;

	case K_ENTER:
		S_LocalSound ("misc/menu2.wav");
		m_return_state = m_state;
		slist_sorted = false;
		key_dest = key_game;
		m_state = m_none;
		Cbuf_AddText ( va ("connect \"%s\"\n", hostcache[slist_cursor].cname) );
		break;

	default:
		break;
	}

}

//=============================================================================
/* Menu Subsystem */


void Menu_MainExitQuake (void);

cmd_t M_ToggleMenu_f_Cmd ("togglemenu", M_ToggleMenu_f);
cmd_t M_Menu_Main_f_Cmd ("menu_main", M_Menu_Main_f);
cmd_t M_Menu_SinglePlayer_f_Cmd ("menu_singleplayer", M_Menu_SinglePlayer_f);
cmd_t M_Menu_Load_f_Cmd ("menu_load", M_Menu_Load_f);
cmd_t M_Menu_Save_f_Cmd ("menu_save", M_Menu_Save_f);
cmd_t M_Menu_MultiPlayer_f_Cmd ("menu_multiplayer", M_Menu_MultiPlayer_f);
cmd_t M_Menu_Setup_f_Cmd ("menu_setup", M_Menu_Setup_f);
cmd_t M_Menu_Options_f_Cmd ("menu_options", M_Menu_Options_f);
cmd_t M_Menu_Keys_f_Cmd ("menu_keys", M_Menu_Keys_f);
cmd_t M_Menu_Video_f_Cmd ("menu_video", M_Menu_Video_f);
cmd_t M_Menu_Help_f_Cmd ("help", M_Menu_Help_f);
cmd_t Menu_MainExitQuake_Cmd ("menu_quit", Menu_MainExitQuake);

void M_Init (void)
{
	// initialize new menu system
	Menu_CommonInit ();
}


void M_Draw (void)
{
	// don't run a draw func if not in the menus or if we don't have a current menu set
	if (m_state == m_none || key_dest != key_menu || !menu_Current) return;

	// draw the appropriate background
	if (scr_con_current)
	{
		Draw_ConsoleBackground (vid.height);
		S_ExtraUpdate ();
	}
	else Draw_FadeScreen ();

	// run the draw func
	menu_Current->Draw ();

	// run the appropriate sound
	switch (menu_soundlevel)
	{
	case m_sound_enter:
		// entering a menu
		S_LocalSound ("misc/menu2.wav");
		break;

	case m_sound_nav:
		// navigating through a menu
		S_LocalSound ("misc/menu1.wav");
		break;

	case m_sound_option:
		// setting a menu's options
		S_LocalSound ("misc/menu3.wav");
		break;

	case m_sound_query:
		// a query that requires user input before continuing
		S_LocalSound ("misc/secret.wav");
		break;

	case m_sound_deny:
		// option denied
		S_LocalSound ("misc/talk.wav");
		break;

	case m_sound_none:
	default:
		// no sound
		break;
	}

	// return to no sounds
	menu_soundlevel = m_sound_none;

	// run an update so that things don't get stuck
	S_ExtraUpdate ();
}


void M_Keydown (int key)
{
	// don't run a key func if we're not in the menus or we don't have a current menu set
	if (m_state == m_none || !menu_Current) return;

	// run the appropriate key func
	menu_Current->Key (key);
}


