

#include "quakedef.h"
#include "menu_common.h"
#include "winquake.h"

// cvars used
extern cvar_t gl_texturedir;
extern cvar_t scr_screenshotformat;
extern cvar_t scr_screenshotdir;
extern cvar_t r_lerporient;
extern cvar_t r_lerpframe;
extern cvar_t r_lerplightstyle;
extern cvar_t host_savedir;
extern cvar_t m_filter;
extern cvar_t m_look;
extern cvar_t m_boost;
extern cvar_t chase_back;
extern cvar_t chase_up;
extern cvar_t chase_right;
extern cvar_t chase_active;
extern cvar_t chase_scale;
extern cvar_t ds_musicdir;

CQMenu menu_Main (NULL, m_main);
CQMenu menu_Singleplayer (&menu_Main, m_other);
CQMenu menu_Save (&menu_Singleplayer, m_save);
CQMenu menu_Load (&menu_Singleplayer, m_load);
CQMenu menu_Multiplayer (&menu_Main, m_multiplayer);
CQMenu menu_TCPIPNewGame (&menu_Multiplayer, m_lanconfig_newgame);
CQMenu menu_TCPIPJoinGame (&menu_Multiplayer, m_lanconfig_joingame);
CQMenu menu_GameConfig (&menu_TCPIPNewGame, m_gameoptions);
CQMenu menu_Setup (&menu_Multiplayer, m_setup);
CQMenu menu_Options (&menu_Main, m_options);
CQMenu menu_Keybindings (&menu_Options, m_keys);
CQMenu menu_Video (&menu_Options, m_video);
CQMenu menu_Help (&menu_Main, m_help);
CQMenu menu_Input (&menu_Options, m_other);
CQMenu menu_Effects (&menu_Options, m_other);
CQMenu menu_ContentDir (&menu_Options, m_other);
CQMenu menu_Chase (&menu_Options, m_other);
CQMenu menu_Search (&menu_TCPIPJoinGame, m_other);
CQMenu menu_SList (&menu_TCPIPJoinGame, m_other);

// for use in various menus
char *SkillNames[] =
{
	"Easy",
	"Normal",
	"Hard",
	"NIGHTMARE",
	NULL
};


// these are temporary until we add a proper maps menu...
char *ID1Maps[] =
{
	// changed these to reflect the actual BSP internal names
	"Introduction (Welcome to Quake)",
	"The Slipgate Complex", "Castle of the Damned", "The Necropolis", "The Grisly Grotto", "Gloom Keep", "The Door To Chthon", "The House of Chthon", "Ziggurat Vertigo",
	"The Installation", "The Ogre Citadel", "The Crypt of Decay", "The Ebon Fortress", "The Wizard's Manse", "The Dismal Oubliette", "The Underearth", 
	"Termination Central", "The Vaults of Zin", "The Tomb of Terror", "Satan's Dark Delight", "The Wind Tunnels", "Chambers of Torment", "The Haunted Halls",
	"The Sewage System", "The Tower of Despair", "The Elder God Shrine", "The Palace of Hate", "Hell's Atrium", "The Pain Maze", "Azure Agony", "The Nameless City",
	"Shub-Niggurath's Pit",
	"Place of Two Deaths", "Claustrophobopolis", "The Abandoned Base", "The Bad Place", "The Cistern", "The Dark Zone",
	NULL
};


char *ID1BSPs[] =
{
	"start",
	"e1m1", "e1m2", "e1m3", "e1m4", "e1m5", "e1m6", "e1m7", "e1m8",
	"e2m1", "e2m2", "e2m3", "e2m4", "e2m5", "e2m6", "e2m7",
	"e3m1", "e3m2", "e3m3", "e3m4", "e3m5", "e3m6", "e3m7",
	"e4m1", "e4m2", "e4m3", "e4m4", "e4m5", "e4m6", "e4m7", "e4m8",
	"end",
	"dm1", "dm2", "dm3", "dm4", "dm5", "dm6",
	NULL
};


char *HipnoMaps[] =
{
	"Command HQ",
	"The Pumping Station", "Storage Facility", "The Lost Mine", "Research Facility", "Military Complex",
	"Ancient Realms", "The Black Cathedral", "The Catacombs", "The Crypt", "Mortum's Keep", "The Gremlin's Domain",
	"Tur Torment", "Pandemonium", "Limbo", "The Gauntlet",
	"Armagon's Lair",
	"The Edge of Oblivion",
	NULL
};


char *HipnoBSPs[] =
{
	"start",
	"hip1m1", "hip1m2", "hip1m3", "hip1m4", "hip1m5",
	"hip2m1", "hip2m2", "hip2m3", "hip2m4", "hip2m5", "hip2m6",
	"hip3m1", "hip3m2", "hip3m3", "hip3m4",
	"hipend",
	"hipdm1",
	NULL
};


char *RogueMaps[] =
{
	"Split Decision", "Deviant's Domain", "Dread Portal", "Judgement Call", "Cave of Death", "Towers of Wrath", "Temple of Pain",
	"Tomb of the Overlord", "Tempus Fugit", "Elemental Fury I", "Elemental Fury II", "Curse of Osiris", "Wizard's Keep",
	"Blood Sacrifice", "Last Bastion", "Source of Evil", "Division of Change", NULL
};


char *RogueBSPs[] =
{
	"start", "r1m1", "r1m2", "r1m3", "r1m4", "r1m5", "r1m6", "r1m7", "r2m1", "r2m2",
	"r2m3", "r2m4", "r2m5", "r2m6", "r2m7", "r2m8", "ctf1", NULL
};


/*
========================================================================================================================

					MAIN MENU

========================================================================================================================
*/

int m_save_demonum;

void Menu_ExitMenus (void)
{
	key_dest = key_game;
	m_state = m_none;
	cls.demonum = m_save_demonum;

	if (cls.demonum != -1 && !cls.demoplayback && cls.state != ca_connected) CL_NextDemo ();
}


void Menu_MainCustomEnter (void)
{
	m_save_demonum = cls.demonum;
	cls.demonum = -1;
}


void Menu_MainExitQuake (void)
{
	// prompt always...
	//if (sv.active)
	if (!Menu_MessageBox ("Are you sure you want to\nExit Quake?\n"))
		return;

	// exit
	key_dest = key_console;
	Host_Quit_f ();
}


void Menu_InitMainMenu (void)
{
	menu_Main.AddOption (new CQMenuBanner ("gfx/ttl_main.lmp"));
	menu_Main.AddOption (new CQMenuTitle ("Select a Submenu"));
	menu_Main.AddOption (new CQMenuSubMenu ("Single Player Menu", &menu_Singleplayer));
	menu_Main.AddOption (new CQMenuSpacer ());
	menu_Main.AddOption (new CQMenuSubMenu ("Multiplayer Menu", &menu_Multiplayer));
	menu_Main.AddOption (new CQMenuSpacer ());
	menu_Main.AddOption (new CQMenuSubMenu ("Configure Game Options", &menu_Options));
	menu_Main.AddOption (new CQMenuSpacer ());
	menu_Main.AddOption (new CQMenuSubMenu ("Help/Ordering", &menu_Help));
	menu_Main.AddOption (new CQMenuSpacer ());
	menu_Main.AddOption (new CQMenuCommand ("Exit Quake", Menu_MainExitQuake));
	menu_Main.AddOption (new CQMenuCustomEnter (Menu_MainCustomEnter));
	menu_Main.AddOption (new CQMenuCustomKey (K_ESCAPE, Menu_ExitMenus));
}



/*
========================================================================================================================

					SINGLE PLAYER MENU

========================================================================================================================
*/


void Menu_SPNewGame (void)
{
	if (sv.active)
		if (!Menu_MessageBox ("Are you sure you want to\nstart a new game?\n"))
			return;

	key_dest = key_game;

	if (sv.active) Cbuf_AddText ("disconnect\n");

	// ensure cvars are appropriate for SP
	Cvar_Set ("deathmatch", 0.0f);
	Cvar_Set ("coop", 0.0f);
	Cvar_Set ("teamplay", 0.0f);

	Cbuf_AddText ("maxplayers 1\n");
	Cbuf_AddText ("map start\n");
}


void Menu_InitSPMenu (void)
{
	menu_Singleplayer.AddOption (new CQMenuBanner ("gfx/ttl_sgl.lmp"));
	menu_Singleplayer.AddOption (new CQMenuTitle ("Single Player Options"));
	menu_Singleplayer.AddOption (new CQMenuCommand ("Start a New Game", Menu_SPNewGame));
	menu_Singleplayer.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Singleplayer.AddOption (new CQMenuSubMenu ("Load a Previously Saved Game", &menu_Load));
	menu_Singleplayer.AddOption (new CQMenuSpacer ());
	menu_Singleplayer.AddOption (new CQMenuSubMenu ("Save the Current Game", &menu_Save));
}


/*
========================================================================================================================

					SAVE/LOAD MENUS

========================================================================================================================
*/

#define MAX_SAVE_DISPLAY	20
int CurrentSaveDisplay = 0;
int NumSaves = 0;
int ActiveItem = 0;

CScrollBoxProvider *SaveLoadScrollbox = NULL;
void Menu_SaveLoadOnDraw (int y, int itemnum);
void Menu_SaveLoadOnHover (int initialy, int y, int itemnum);
void Menu_SaveLoadOnEnter (int itemnum);


typedef struct save_game_info_s
{
	char mapname[SAVEGAME_COMMENT_LENGTH + 1];
	char kills[64];
	int skill;
	char time[64];
	char secrets[64];
	char savetime[64];
	char filename[MAX_OSPATH];
} save_game_info_t;


void GetSaveInfo (FILE *f, char *filename, save_game_info_t *si)
{
	int		j;
	char	name[MAX_OSPATH];
	int		version;
	float	dummy;
	int		stuff_total;
	int		stuff_done;
	SYSTEMTIME st;

	// copy the file name
	strcpy (si->filename, filename);

	// read the header comment
	fscanf (f, "%79s\n", name);

	// change _ to space
	for (j = 0; j < SAVEGAME_COMMENT_LENGTH; j++)
		if (name[j] == '_') name[j] = ' ';

	// strip out "kills"
	// (note: per Host_SavegameComment kills always starts at position 22)
	name[22] = 0;

	// copy in the map name
	strncpy (si->mapname, name, SAVEGAME_COMMENT_LENGTH);

	// parse out number of kills
	// (note: per Host_SavegameComment number of kills always starts at position 28)
	sscanf (&name[28], "%i/%i", &stuff_done, &stuff_total);

	// copy to kills
	// note: we did it this way so that we could strip out the whitespace from it.
	// kills numbers are written as %3i, but parsing the array would break if there was > 999 of either
	sprintf (si->kills, "%i/%i", stuff_done, stuff_total);

	// read and skip over spawn parms
	for (j = 0; j < NUM_SPAWN_PARMS; j++)
		fscanf (f, "%f\n", &dummy);

	// read skill as a float, and convert to int
	fscanf (f, "%f\n", &dummy);
	si->skill = (int) (dummy + 0.1);

	// sanity check (in case the save is manually hacked...)
	if (si->skill > 3) si->skill = 3;
	if (si->skill < 0) si->skill = 0;

	// read bsp mapname
	fscanf (f, "%s\n", name);

	// now read time for real as a float
	fscanf (f, "%f\n", &dummy);

	// convert dummy time to real time
	sprintf (si->time, "%02i:%02i", ((int) dummy) / 60, (int) dummy - (((int) dummy) / 60) * 60);

	// initially nothing (differentiate from 0 as there may be genuinely 0)
	stuff_total = -1;
	stuff_done = -1;

	// now parse the rest of the file looking for the other stuff we need
	while (1)
	{
		// scan into secrets as a dummy
		fscanf (f, "%s", si->secrets);

		if (!strcmp (si->secrets, "\"stuff_total\"") && stuff_total == -1)
		{
			// scan again to get the data
			fscanf (f, "%s", si->secrets);

			// these are stored as floats, so we terminate at the .
			for (j = 0; ; j++)
			{
				if (si->secrets[j] == '.')
				{
					si->secrets[j] = 0;
					break;
				}
			}

			stuff_total = Q_atoi (&si->secrets[1]);
		}

		if (!strcmp (si->secrets, "\"stuff_done\"") && stuff_done == -1)
		{
			// scan again to get the data
			fscanf (f, "%s", si->secrets);

			// these are stored as floats, so we terminate at the .
			for (j = 0; ; j++)
			{
				if (si->secrets[j] == '.')
				{
					si->secrets[j] = 0;
					break;
				}
			}

			stuff_done = Q_atoi (&si->secrets[1]);
		}

		// eof
		if (feof (f)) break;

		// got them all
		if (!(stuff_total == -1 || stuff_done == -1)) break;
	}

	if (stuff_total == -1) stuff_total = 0;
	if (stuff_done == -1) stuff_done = 0;

	if (stuff_done > stuff_total) stuff_done = stuff_total;

	sprintf (si->secrets, "%i/%i", stuff_done, stuff_total);

	fclose (f);

	char name2[256];

	// set up the file name for opening
	sprintf (name2, "%s/%s/%s.sav", com_gamedir, host_savedir.string, filename);

	// open it again (isn't the Windows API beautiful?)
	HANDLE hFile = CreateFile (name2, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if (!hFile) return;

	BY_HANDLE_FILE_INFORMATION FileInfo;

	// retrieve the info
	GetFileInformationByHandle (hFile, &FileInfo);

	// use the last write time so that it's always valid for the last save
	// there's no FileTimeToLocalTime - grrrrr...!
	FILETIME lft;
	FileTimeToLocalFileTime (&FileInfo.ftLastWriteTime, &lft);
	FileTimeToSystemTime (&lft, &st);

	sprintf (si->savetime, "%04i-%02i-%02i %02i:%02i", st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute);

	// close the file handle
	CloseHandle (hFile);
}

class CSaveInfo
{
public:
	CSaveInfo (FILE *f, char *filename)
	{
		GetSaveInfo (f, filename, &this->SaveInfo);
		NumSaves++;
	}

	CSaveInfo (void)
	{
		strcpy (this->SaveInfo.mapname, "<< New SaveGame >>");
		NumSaves++;
	}

	~CSaveInfo ()
	{
		// cascade destructors along the list
		if (this->Next)
		{
			delete this->Next;
			this->Next = NULL;
		}
	}

	CSaveInfo *Next;
	save_game_info_t SaveInfo;
};


CSaveInfo *SaveInfoList = NULL;
CSaveInfo **SaveInfoArray = NULL;


void Menu_SaveLoadAddSave (WIN32_FIND_DATA *savefile)
{
	// not interested in these types
	if (savefile->dwFileAttributes & FILE_ATTRIBUTE_COMPRESSED) return;
	if (savefile->dwFileAttributes & FILE_ATTRIBUTE_ENCRYPTED) return;
	if (savefile->dwFileAttributes & FILE_ATTRIBUTE_OFFLINE) return;
	if (savefile->dwFileAttributes & FILE_ATTRIBUTE_SYSTEM) return;
	if (savefile->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) return;

	// attempt to open it
	FILE *f = fopen (va ("%s\\save\\%s", com_gamedir, savefile->cFileName), "r");
	int version;

	// failed to open (we don't expect this to happen)
	if (!f) return;

	// check the version
	fscanf (f, "%i\n", &version);

	// check the version
	if (version != SAVEGAME_VERSION) return;

	if (!SaveInfoList)
	{
		// first item begins the list
		SaveInfoList = new CSaveInfo (f, savefile->cFileName);
		SaveInfoList->Next = NULL;
		return;
	}

	// add to the end of the list
	for (CSaveInfo *si = SaveInfoList; si; si = si->Next)
	{
		// find the item before the last
		if (!si->Next)
		{
			// link in, terminate the list, and get out
			si->Next = new CSaveInfo (f, savefile->cFileName);
			si->Next->Next = NULL;
			return;
		}
	}
}


void Menu_SaveLoadScanSaves (void)
{
	// destroy the previous list
	if (SaveInfoList)
	{
		delete SaveInfoList;
		SaveInfoList = NULL;
	}

	if (SaveInfoArray)
	{
		free (SaveInfoArray);
		SaveInfoArray = NULL;
	}

	if (SaveLoadScrollbox)
	{
		delete SaveLoadScrollbox;
		SaveLoadScrollbox = NULL;
	}

	// set display params
	CurrentSaveDisplay = 0;
	NumSaves = 0;
	ActiveItem = 0;

	// make the save directory
	COM_CheckContentDirectory (&host_savedir, true);

	if (m_state == m_save)
	{
		// create a first item for the new savegame
		SaveInfoList = new CSaveInfo ();
		SaveInfoList->Next = NULL;
	}

	WIN32_FIND_DATA FindFileData;
	HANDLE hFind = INVALID_HANDLE_VALUE;

	// look for a file
	hFind = FindFirstFile (va ("%s\\%s\\*.sav", com_gamedir, host_savedir.string), &FindFileData);

	if (hFind == INVALID_HANDLE_VALUE)
	{
		// found no files
		FindClose (hFind);
	}
	else
	{
		// add the first
		Menu_SaveLoadAddSave (&FindFileData);

		// add the rest
		while (FindNextFile (hFind, &FindFileData))
			Menu_SaveLoadAddSave (&FindFileData);

		// close the finder
		FindClose (hFind);
	}

	// now put them all in an array for easy access
	SaveInfoArray = (CSaveInfo **) malloc (sizeof (CSaveInfo *) * NumSaves);
	int SaveIndex = 0;

	for (CSaveInfo *si = SaveInfoList; si; si = si->Next)
	{
		SaveInfoArray[SaveIndex] = si;
		SaveIndex++;
	}

	// create the scrollbox provider
	SaveLoadScrollbox = new CScrollBoxProvider (NumSaves, MAX_SAVE_DISPLAY, 22);
	SaveLoadScrollbox->SetDrawItemCallback (Menu_SaveLoadOnDraw);
	SaveLoadScrollbox->SetHoverItemCallback (Menu_SaveLoadOnHover);
	SaveLoadScrollbox->SetEnterItemCallback (Menu_SaveLoadOnEnter);
}


void Menu_SaveCustomEnter (void)
{
	if (!sv.active || cl.intermission || svs.maxclients != 1 || cl.stats[STAT_HEALTH] <= 0)
	{
		// can't save
		menu_soundlevel = m_sound_deny;
		return;
	}

	// scan the saves
	Menu_SaveLoadScanSaves ();
}


void Menu_LoadCustomEnter (void)
{
	// scan the saves
	Menu_SaveLoadScanSaves ();

	// check number of saves - if 0, sound a deny
	if (!SaveInfoList) menu_soundlevel = m_sound_deny;
}


void Draw_Mapshot (char *name, int x, int y, int w, int h);

void Menu_SaveLoadOnHover (int initialy, int y, int itemnum)
{
	// highlight bar
	Menu_HighlightBar (-174, y, 172);

	// mapshot box
	Draw_TextBox ((vid.width - 320) / 2 + 200, initialy, 128, 128);

	if (m_state == m_save && itemnum == 0)
	{
		// new savegame
		Draw_Mapshot (NULL, (vid.width - 320) / 2 + 208, initialy + 8, 128, 128);
		M_PrintWhite (220, initialy + 145, "Current Stats");
		M_Print (218, initialy + 160, DIVIDER_LINE);
		M_Print (220, initialy + 175, va ("Kills:   %i/%i", cl.stats[STAT_MONSTERS], cl.stats[STAT_TOTALMONSTERS]));
		M_Print (220, initialy + 187, va ("Secrets: %i/%i", cl.stats[STAT_SECRETS], cl.stats[STAT_TOTALSECRETS]));
		M_Print (220, initialy + 199, va ("Skill:   %s", SkillNames[(int) skill.value]));
		M_Print (220, initialy + 211, va ("Time:    %02i:%02i", (int) (cl.time / 60), (int) (cl.time - (int) (cl.time / 60) * 60)));
	}
	else
	{
		// existing save game
		Draw_Mapshot (va ("save/%s", SaveInfoArray[itemnum]->SaveInfo.filename), (vid.width - 320) / 2 + 208, initialy + 8, 128, 128);
		M_PrintWhite (220, initialy + 145, "Savegame Info");
		M_Print (218, initialy + 160, DIVIDER_LINE);
		M_Print (220, initialy + 175, va ("Kills:   %s", SaveInfoArray[itemnum]->SaveInfo.kills));
		M_Print (220, initialy + 187, va ("Secrets: %s", SaveInfoArray[itemnum]->SaveInfo.secrets));
		M_Print (220, initialy + 199, va ("Skill:   %s", SkillNames[SaveInfoArray[itemnum]->SaveInfo.skill]));
		M_Print (220, initialy + 211, va ("Time:    %s", SaveInfoArray[itemnum]->SaveInfo.time));
	}

	// FileTimeToSystemTime seems broken so this is removed until i fix it :(
	// M_Print (200, 190, va ("%s", SaveInfoArray[itemnum]->SaveInfo.savetime));
}


void Menu_SaveLoadOnDraw (int y, int itemnum)
{
	// draw the item
	M_Print (-8, y, SaveInfoArray[itemnum]->SaveInfo.mapname);
}


void Menu_SaveLoadOnEnter (int itemnum)
{
	// ensure the folder exists
	COM_CheckContentDirectory (&host_savedir, true);

	if (m_state == m_save)
	{
		if (itemnum == 0)
		{
			int i;

			// generate a new save name
			for (i = 0; i < 99999; i++)
			{
				FILE *f = fopen (va ("%s/%s/save_%05i.sav", com_gamedir, host_savedir.string, i), "r");

				if (!f)
				{
					// save to this one
					Cbuf_AddText (va ("save save_%05i\n", i));
					break;
				}

				fclose (f);
			}

			if (i == 100000)
				Con_Printf ("Menu_SaveLoadOnEnter: Failed to find a free savegame slot\n");

			// exit menus
			m_state = m_none;
			key_dest = key_game;
			return;
		}

		// overwrite save
		if (Menu_MessageBox (va ("%s\n\nAre you sure that you want to overwrite this save?", SaveInfoArray[itemnum]->SaveInfo.filename)))
			Cbuf_AddText (va ("save %s\n", SaveInfoArray[itemnum]->SaveInfo.filename));

		// exit menus
		m_state = m_none;
		key_dest = key_game;

		return;
	}

	// exit menus
	m_state = m_none;
	key_dest = key_game;

	// Host_Loadgame_f can't bring up the loading plaque because too much
	// stack space has been used, so do it now
	SCR_BeginLoadingPlaque ();

	// to do - check if client status has changed and if so, display a warning
	Cbuf_AddText (va ("load %s\n", SaveInfoArray[itemnum]->SaveInfo.filename));
}


int Menu_SaveLoadCustomDraw (int y)
{
	// check for validity, report and get out
	if (m_state == m_save)
	{
		// false if a save cannot be done at this time
		bool validsave = true;

		if (!sv.active)
		{
			M_PrintCenter (y + 10, "You cannot save with an inactive local server");
			validsave = false;
		}

		if (cl.intermission)
		{
			M_PrintCenter (y + 10, "You cannot save during an intermission");
			validsave = false;
		}

		if (svs.maxclients != 1)
		{
			M_PrintCenter (y + 10, "You cannot save a multiplayer game");
			validsave = false;
		}

		if (cl.stats[STAT_HEALTH] <= 0)
		{
			M_PrintCenter (y + 10, "You cannot save with a dead player");
			validsave = false;
		}

		if (!validsave)
		{
			M_PrintCenter (y + 30, "Press ESC to Exit this Menu");
			return y + 40;
		}
	}
	else
	{
		if (!SaveInfoList)
		{
			M_PrintCenter (y + 10, "There are no Saved Games to Load from");
			M_PrintCenter (y + 30, "Press ESC to Exit this Menu");
			return y + 40;
		}
	}

	y = SaveLoadScrollbox->DrawItems ((vid.width - 320) / 2 - 24, y);

	return y;
}


void Menu_SaveLoadCustomKey (int key)
{
	if (!SaveInfoList) return;
	if (!SaveLoadScrollbox) return;

	switch (key)
	{
	case K_DOWNARROW:
	case K_UPARROW:
		menu_soundlevel = m_sound_nav;

	case K_ENTER:
		SaveLoadScrollbox->KeyFunc (key);
		break;
	}
}


void Menu_InitSaveLoadMenu (void)
{
	menu_Save.AddOption (new CQMenuBanner ("gfx/p_save.lmp"));
	menu_Save.AddOption (new CQMenuTitle ("Save the Current Game"));
	menu_Save.AddOption (new CQMenuCustomEnter (Menu_SaveCustomEnter));
	menu_Save.AddOption (new CQMenuCustomDraw (Menu_SaveLoadCustomDraw));
	menu_Save.AddOption (new CQMenuCustomKey (K_UPARROW, Menu_SaveLoadCustomKey));
	menu_Save.AddOption (new CQMenuCustomKey (K_DOWNARROW, Menu_SaveLoadCustomKey));
	menu_Save.AddOption (new CQMenuCustomKey (K_ENTER, Menu_SaveLoadCustomKey));

	menu_Load.AddOption (new CQMenuBanner ("gfx/p_load.lmp"));
	menu_Load.AddOption (new CQMenuTitle ("Load a Previously Saved Game"));
	menu_Load.AddOption (new CQMenuCustomEnter (Menu_LoadCustomEnter));
	menu_Load.AddOption (new CQMenuCustomDraw (Menu_SaveLoadCustomDraw));
	menu_Load.AddOption (new CQMenuCustomKey (K_UPARROW, Menu_SaveLoadCustomKey));
	menu_Load.AddOption (new CQMenuCustomKey (K_DOWNARROW, Menu_SaveLoadCustomKey));
	menu_Load.AddOption (new CQMenuCustomKey (K_ENTER, Menu_SaveLoadCustomKey));
}


/*
========================================================================================================================

					MULTI PLAYER MENU

========================================================================================================================
*/

int Menu_MultiplayerCustomDraw (int y)
{
	if (serialAvailable || ipxAvailable || tcpipAvailable) return y;

	M_PrintCenterWhite (vid.height - 80, "No Communications Available");

	return y;
}


// dummy cvars to make textboxes not overwrite the real ones
cvar_t dummy_name;
cvar_t dummy_hostname;

// player colours
int setup_shirt, setup_oldshirt;
int setup_pants, setup_oldpants;

void Menu_SetupApplyFunc (void)
{
	// execute the commands
	Cbuf_AddText (va ("name \"%s\"\n", dummy_name.string));
	Cvar_Set ("hostname", dummy_hostname.string);

	if (setup_shirt != setup_oldshirt || setup_pants != setup_oldpants)
		Cbuf_AddText (va ("color %i %i\n", setup_shirt, setup_pants));

	// return to the multiplayer menu
	menu_Multiplayer.EnterMenu ();
}


void Menu_SetupCustomEnter (void)
{
	// copy cvars out
	strncpy (dummy_name.string, cl_name.string, 1023);
	strncpy (dummy_hostname.string, hostname.string, 1023);

	setup_shirt = setup_oldshirt = ((int) cl_color.value) >> 4;
	setup_pants = setup_oldpants = ((int) cl_color.value) & 15;
}


byte identityTable[256];
byte translationTable[256];

void M_BuildTranslationTable (int top, int bottom)
{
	int		j;
	byte	*dest, *source;

	for (j = 0; j < 256; j++)
		identityTable[j] = j;

	dest = translationTable;
	source = identityTable;
	memcpy (dest, source, 256);

	if (top < 128)	// the artists made some backwards ranges.  sigh.
		memcpy (dest + TOP_RANGE, source + top, 16);
	else
		for (j = 0; j < 16; j++)
			dest[TOP_RANGE + j] = source[top + 15 - j];

	if (bottom < 128)
		memcpy (dest + BOTTOM_RANGE, source + bottom, 16);
	else
		for (j = 0; j < 16; j++)
			dest[BOTTOM_RANGE + j] = source[bottom + 15 - j];
}


int Menu_SetupCustomDraw (int y)
{
	qpic_t *p;
	int newy = y;

	p = Draw_CachePic ("gfx/bigbox.lmp");
	Draw_Pic ((vid.width - p->width) / 2, y, p);
	newy += p->height + 5;

	p = Draw_CachePic ("gfx/menuplyr.lmp");
	M_BuildTranslationTable (setup_shirt * 16, setup_pants * 16);
	Draw_PicTranslate ((vid.width - p->width) / 2, y + 8, p, translationTable);

	return newy;
}


int Menu_TCPIPCustomDraw (int y)
{
	M_Print (148 - strlen ("IP Address") * 8, y, "IP Address");
	M_PrintWhite (172, y, my_tcpip_address);

	y += 15;

	M_Print (148 - strlen ("Host Name") * 8, y, "Host Name");
	M_PrintWhite (172, y, hostname.string);

	y += 15;

	if (serialAvailable || ipxAvailable || tcpipAvailable) return y;

	M_PrintCenterWhite (vid.height - 80, "No Communications Available");

	return y;
}


cvar_t dummy_port;
cvar_t dummy_remoteip;

void Menu_TCPIPCustomEnter (void)
{
	// copy out the port
	sprintf (dummy_port.string, "%i", DEFAULTnet_hostport);
	dummy_remoteip.string[0] = 0;
}


void Menu_TCPIPPostConfigCommon (void)
{
	// stop all demos
	Cbuf_AddText ("stopdemo\n");

	// retrieve the port
	int l =  Q_atoi (dummy_port.string);

	// store it out
	if (l > 65535)
		net_hostport = DEFAULTnet_hostport;
	else net_hostport = l;
}


void Menu_TCPIPContinueToGameOptions (void)
{
	// complete net config
	Menu_TCPIPPostConfigCommon ();

	// fire the game options menu entry function
	menu_GameConfig.EnterMenu ();
}


cvar_t dummy_maxplayers;
cvar_t dummy_timelimit;
cvar_t dummy_fraglimit;
int dummy_skill;
int dummy_teamplay;

int map_number = 0;
char **MapNames = ID1Maps;
char **MapBSPs = ID1BSPs;

char *ID1Teamplay[] = {"Off", "No Friendly Fire", "Friendly Fire", NULL};
char *RogueTeamplay[] = {"Off", "No Friendly Fire", "Friendly Fire", "Tag", "Capture the Flag", "One Flag CTF", "Three Team CTF", NULL};
char **ActiveTeamplay = ID1Teamplay;
char *GameTypes[] = {"Cooperative", "Deathmatch", NULL};
int gametype_number = 0;


int Menu_GameConfigCustomDraw (int y)
{
	if ((int) dummy_maxplayers.value == svs.maxclientslimit)
	{
		if (svs.maxclientslimit == MAX_SCOREBOARD)
			M_PrintCenter (vid.height - 75, va ("Cannot select more than %i players", svs.maxclientslimit));
		else
		{
			M_PrintCenter (vid.height - 75, va ("Running with more than %i players", svs.maxclientslimit));
			M_PrintCenter (vid.height - 65, "Requires using command-line parameters");
			M_PrintCenter (vid.height - 50, "Please see TECHINFO.TXT");
		}
	}

	return y;
}


void Menu_GameConfigCustomEnter (void)
{
	// copy out cvars to prevent in-place editing of values screwing with current settings before we commit them
	// especially as many of these are broadcast cvars...
	dummy_maxplayers.value = svs.maxclients;
	dummy_fraglimit.value = fraglimit.value;
	dummy_timelimit.value = timelimit.value;
	dummy_skill = (int) (skill.value + 0.5);
	dummy_teamplay = (int) teamplay.value;

	// bound values so we don't crash or otherwise behave unexpectedly
	if (dummy_maxplayers.value < 2) dummy_maxplayers.value = 2; else if (dummy_maxplayers.value > svs.maxclientslimit) dummy_maxplayers.value = svs.maxclientslimit;
	if (dummy_fraglimit.value < 0) dummy_fraglimit.value = 0; else if (dummy_fraglimit.value > 100) dummy_fraglimit.value = 100;
	if (dummy_timelimit.value < 0) dummy_timelimit.value = 0; else if (dummy_timelimit.value > 60) dummy_timelimit.value = 60;
	if (dummy_skill < 0) dummy_skill = 0; else if (dummy_skill > 3) dummy_skill = 3;

	if (rogue) {if (dummy_teamplay < 0) dummy_teamplay = 0; else if (dummy_teamplay > 6) dummy_teamplay = 6;}
	else {if (dummy_teamplay < 0) dummy_teamplay = 0; else if (dummy_teamplay > 2) dummy_teamplay = 2;}

	if (coop.value)
		gametype_number = 0;
	else gametype_number = 1;
}


void Menu_GameConfigBeginGame (void)
{
	// disconnet any current servers
	if (sv.active) Cbuf_AddText ("disconnect\n");

	// so host_netport will be re-examined
	Cbuf_AddText ("listen 0\n");
	Cbuf_AddText (va ("maxplayers %u\n", (int) dummy_maxplayers.value));
	SCR_BeginLoadingPlaque ();

	// copy cvar values back now that we're committed
	Cvar_Set (&timelimit, dummy_timelimit.value);
	Cvar_Set (&fraglimit, dummy_fraglimit.value);
	Cvar_Set (&skill, (float) dummy_skill);
	Cvar_Set (&teamplay, (float) dummy_teamplay);

	// set gametype
	if (gametype_number)
	{
		Cvar_Set (&coop, 0.0f);
		Cvar_Set (&deathmatch, 1.0f);
	}
	else
	{
		Cvar_Set (&coop, 1.0f);
		Cvar_Set (&deathmatch, 0.0f);
	}

	// load the correct map
	Cbuf_AddText (va ("map %s\n", MapBSPs[map_number]));
}


bool Menu_SearchComplete = false;
double Menu_SearchCompleteTime;

void Menu_SearchCustomEnter (void)
{
	slistSilent = true;
	slistLocal = false;
	Menu_SearchComplete = false;
	NET_Slist_f ();
}


int Menu_SearchCustomDraw (int y)
{
	if (slistInProgress)
	{
		M_PrintCenterWhite (y + 50, "...Searching...");
		NET_Poll ();
		return y;
	}

	if (!Menu_SearchComplete)
	{
		Menu_SearchComplete = true;
		Menu_SearchCompleteTime = realtime;
	}

	if (hostCacheCount)
	{
		menu_SList.EnterMenu ();
		return y;
	}

	M_PrintCenterWhite (y + 50, "No Quake Servers found");

	if ((realtime - Menu_SearchCompleteTime) < 3.0) return y;

	menu_TCPIPJoinGame.EnterMenu ();

	return y;
}


CScrollBoxProvider *MenuSListScrollbox = NULL;

void Menu_SListOnDraw (int y, int itemnum)
{
	// draw the item
	M_Print (-8, y, hostcache[itemnum].name);
}


void Menu_SListOnHover (int initialy, int y, int itemnum)
{
	// highlight bar
	Menu_HighlightBar (-174, y, 172);

	// mapshot box
	Draw_TextBox ((vid.width - 320) / 2 + 200, initialy, 128, 128);

	// rest of info
	Draw_Mapshot (NULL, (vid.width - 320) / 2 + 208, initialy + 8, 128, 128);
	M_PrintWhite (228, initialy + 145, "Server Info");
	M_Print (218, initialy + 160, DIVIDER_LINE);

	hostcache_t *hc = &hostcache[itemnum];

	char hostipconfig[32];
	char *hostport;

	strcpy (hostipconfig, hc->cname);

	for (int i = strlen (hostipconfig) - 1; i; i--)
	{
		if (hostipconfig[i] == ':')
		{
			hostport = &hostipconfig[i + 1];
			hostipconfig[i] = 0;
			break;
		}
	}

	M_Print (212, initialy + 175, va ("IP Addr: %s", hostipconfig));
	M_Print (212, initialy + 185, va ("Port:    %s", hostport));
	M_Print (218, initialy + 200, DIVIDER_LINE);
	M_Print (212, initialy + 215, va ("Map:     %s", hc->map));
	M_Print (212, initialy + 225, va ("Players: %i of %i", hc->users, hc->maxusers));
}


char m_return_reason [32];
CQMenu *Menu_NetReturn = NULL;

float Net_ErrorReturnTime = 0;

void NET_MenuReturn (void)
{
	if (Menu_NetReturn)
	{
		Net_ErrorReturnTime = realtime + 5;
		Menu_NetReturn->EnterMenu ();
	}

	Menu_NetReturn = NULL;
}


int Menu_NetErrorCustomDraw (int y)
{
	if (Net_ErrorReturnTime < realtime) return y;

	Draw_TextBox ((vid.width / 2) - 175, 80, 334, 90);
	M_PrintCenterWhite (100, "Connection Error");
	M_PrintCenter (115, DIVIDER_LINE);
	M_PrintCenter (130, "Failed to connect to remote server");
	M_PrintCenter (145, va ("The reason given was:"));
	M_PrintCenter (160, va ("\"%s\"", m_return_reason));

	return y;
}


void Menu_MultiplayerConnectToGame (CQMenu *returnmenu, char *remote)
{
	key_dest = key_game;
	m_state = m_none;
	Menu_NetReturn = returnmenu;
	m_return_reason[0] = 0;
	Cbuf_AddText (va ("connect \"%s\"\n", remote));
}


void Menu_SListOnEnter (int itemnum)
{
	// join this game
	Menu_MultiplayerConnectToGame (&menu_SList, hostcache[itemnum].cname);
}


void Menu_SListCustomKey (int key)
{
	if (!MenuSListScrollbox) return;

	switch (key)
	{
	case K_DOWNARROW:
	case K_UPARROW:
		menu_soundlevel = m_sound_nav;

	case K_ENTER:
		MenuSListScrollbox->KeyFunc (key);
		break;
	}
}


void Menu_SListCustomEnter (void)
{
	// sort the server list
	if (hostCacheCount > 1)
	{
		int	i, j;
		hostcache_t temp;

		for (i = 0; i < hostCacheCount; i++)
		{
			for (j = i + 1; j < hostCacheCount; j++)
			{
				if (strcmp (hostcache[j].name, hostcache[i].name) < 0)
				{
					Q_memcpy (&temp, &hostcache[j], sizeof (hostcache_t));
					Q_memcpy (&hostcache[j], &hostcache[i], sizeof (hostcache_t));
					Q_memcpy (&hostcache[i], &temp, sizeof (hostcache_t));
				}
			}
		}
	}

	// create a scrollbox for it
	if (MenuSListScrollbox) delete MenuSListScrollbox;
	MenuSListScrollbox = new CScrollBoxProvider (hostCacheCount, 20, 22);
	MenuSListScrollbox->SetDrawItemCallback (Menu_SListOnDraw);
	MenuSListScrollbox->SetHoverItemCallback (Menu_SListOnHover);
	MenuSListScrollbox->SetEnterItemCallback (Menu_SListOnEnter);
}


int Menu_SListCustomDraw (int y)
{
	if (!MenuSListScrollbox) return y;

	// draw through the scrollbox provider
	MenuSListScrollbox->DrawItems ((vid.width - 320) / 2 - 24, y);

	return y;
}


void Menu_TCPIPJoinGameCommand (void)
{
	// connect to this game
	Menu_MultiplayerConnectToGame (&menu_TCPIPJoinGame, dummy_remoteip.string);
}


void Menu_InitMultiplayerMenu (void)
{
	// top level menu
	menu_Multiplayer.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_Multiplayer.AddOption (new CQMenuTitle ("Configure Multiplayer Game Options"));
	menu_Multiplayer.AddOption (new CQMenuSubMenu ("Join an Existing Multiplayer Game", &menu_TCPIPJoinGame));
	menu_Multiplayer.AddOption (new CQMenuSpacer ());
	menu_Multiplayer.AddOption (new CQMenuSubMenu ("Start a New Multiplayer Game", &menu_TCPIPNewGame));
	menu_Multiplayer.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Multiplayer.AddOption (new CQMenuSubMenu ("Setup Player Options", &menu_Setup));
	menu_Multiplayer.AddOption (new CQMenuCustomDraw (Menu_MultiplayerCustomDraw));

	// setup menu
	menu_Setup.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_Setup.AddOption (new CQMenuTitle ("Setup Player Options"));
	menu_Setup.AddOption (new CQMenuCustomEnter (Menu_SetupCustomEnter));
	menu_Setup.AddOption (new CQMenuCvarTextbox ("Player Name", &dummy_name));
	menu_Setup.AddOption (new CQMenuCvarTextbox ("Host Name", &dummy_hostname));
	menu_Setup.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Setup.AddOption (new CQMenuColourBar ("Shirt Colour", &setup_shirt));
	menu_Setup.AddOption (new CQMenuCustomDraw (Menu_SetupCustomDraw));
	menu_Setup.AddOption (new CQMenuColourBar ("Pants Colour", &setup_pants));
	menu_Setup.AddApplyCommand ("Apply Changes", Menu_SetupApplyFunc);

	// start a new game
	menu_TCPIPNewGame.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_TCPIPNewGame.AddOption (new CQMenuTitle ("Configure TCP/IP Options - New Game"));
	menu_TCPIPNewGame.AddOption (new CQMenuCustomEnter (Menu_TCPIPCustomEnter));
	menu_TCPIPNewGame.AddOption (new CQMenuCustomDraw (Menu_TCPIPCustomDraw));
	menu_TCPIPNewGame.AddOption (new CQMenuCvarTextbox ("Port", &dummy_port, true));
	menu_TCPIPNewGame.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_TCPIPNewGame.AddOption (new CQMenuCommand ("Continue to Game Options", Menu_TCPIPContinueToGameOptions));

	// join an existing game
	menu_TCPIPJoinGame.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_TCPIPJoinGame.AddOption (new CQMenuTitle ("Configure TCP/IP Options - Join a Game"));
	menu_TCPIPJoinGame.AddOption (new CQMenuCustomEnter (Menu_TCPIPCustomEnter));
	menu_TCPIPJoinGame.AddOption (new CQMenuCustomDraw (Menu_TCPIPCustomDraw));
	menu_TCPIPJoinGame.AddOption (new CQMenuCvarTextbox ("Port", &dummy_port, true));
	menu_TCPIPJoinGame.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_TCPIPJoinGame.AddOption (new CQMenuSubMenu ("Search for Local Games", &menu_Search));
	menu_TCPIPJoinGame.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_TCPIPJoinGame.AddOption (new CQMenuCvarTextbox ("Remote Server", &dummy_remoteip));
	menu_TCPIPJoinGame.AddOption (new CQMenuCommand ("Join this Game", Menu_TCPIPJoinGameCommand));
	menu_TCPIPJoinGame.AddOption (new CQMenuCustomDraw (Menu_NetErrorCustomDraw));

	// game options
	if (rogue) ActiveTeamplay = RogueTeamplay;
	menu_GameConfig.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_GameConfig.AddOption (new CQMenuTitle ("Game Configuration Options"));
	menu_GameConfig.AddOption (new CQMenuCustomEnter (Menu_GameConfigCustomEnter));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Game Type", &gametype_number, GameTypes));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Teamplay", &dummy_teamplay, ActiveTeamplay));
	menu_GameConfig.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Maximum Players", &dummy_maxplayers, 2, svs.maxclientslimit, 1, NULL, NULL));
	menu_GameConfig.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Skill", &dummy_skill, SkillNames));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Frag Limit", &dummy_fraglimit, 0, 100, 10, "None", "Frags"));
	menu_GameConfig.AddOption (new CQMenuSpinControl ("Time Limit", &dummy_timelimit, 0, 60, 5, "None", "Minutes"));
	menu_GameConfig.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_GameConfig.AddOption (new CQMenuSpacer ("Select a Map"));
	menu_GameConfig.AddOption (new CQMenuCustomDraw (Menu_GameConfigCustomDraw));

	// set up the correct maps menu
	if (hipnotic)
	{
		MapNames = HipnoMaps;
		MapBSPs = HipnoBSPs;
	}
	else if (rogue)
	{
		MapNames = RogueMaps;
		MapBSPs = RogueBSPs;
	}

	// and add it
	menu_GameConfig.AddOption (new CQMenuSpinControl (NULL, &map_number, MapNames));
	menu_GameConfig.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_GameConfig.AddOption (new CQMenuCommand ("Begin New Multiplayer Game", Menu_GameConfigBeginGame));

	// search for games
	menu_Search.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_Search.AddOption (new CQMenuTitle ("Search for Local Games"));
	menu_Search.AddOption (new CQMenuCustomEnter (Menu_SearchCustomEnter));
	menu_Search.AddOption (new CQMenuCustomDraw (Menu_SearchCustomDraw));

	// server list
	menu_SList.AddOption (new CQMenuBanner ("gfx/p_multi.lmp"));
	menu_SList.AddOption (new CQMenuTitle ("Local Quake Server List"));
	menu_SList.AddOption (new CQMenuCustomEnter (Menu_SListCustomEnter));
	menu_SList.AddOption (new CQMenuCustomDraw (Menu_SListCustomDraw));
	menu_SList.AddOption (new CQMenuCustomDraw (Menu_NetErrorCustomDraw));
	menu_SList.AddOption (new CQMenuCustomKey (K_UPARROW, Menu_SListCustomKey));
	menu_SList.AddOption (new CQMenuCustomKey (K_DOWNARROW, Menu_SListCustomKey));
	menu_SList.AddOption (new CQMenuCustomKey (K_ENTER, Menu_SListCustomKey));
}


/*
========================================================================================================================

					OPTIONS MENU

========================================================================================================================
*/


// commands
void Host_WriteConfiguration (void);

void Menu_OptionsResetToDefaults (void)
{
	Cbuf_AddText ("exec default.cfg\n");
}


void Menu_OptionsGoToConsole (void)
{
	m_state = m_none;
	Con_ToggleConsole_f ();
}


char *key_bindnames[][2] =
{
	{"+attack", 		"Attack"},
	{"impulse 10", 		"Next Weapon"},
	{"impulse 12", 		"Previous Weapon"},
	{"+jump", 			"Jump/Swim up"},
	{"+forward", 		"Walk Forward"},
	{"+back", 			"Backpedal"},
	{"+left", 			"Turn Left"},
	{"+right", 			"Turn Right"},
	{"+speed", 			"Run"},
	{"+moveleft", 		"Step Left"},
	{"+moveright", 		"Step Right"},
	{"+strafe", 		"Sidestep"},
	{"+lookup", 		"Look Up"},
	{"+lookdown", 		"Look Down"},
	{"centerview", 		"Center View"},
	{"+mlook", 			"Mouse Look"},
	{"+klook", 			"Keyboard Look"},
	{"+moveup",			"Swim Up"},
	{"+movedown",		"Swim Down"}
};


bool keybind_grab = false;

#define	NUMBINDINGS		(sizeof (key_bindnames) / sizeof (key_bindnames[0]))

int bind_cursor = 0;


void Menu_FindKeybindingsForAction (char *action, int *twokeys)
{
	char	*b;

	twokeys[0] = twokeys[1] = -1;
	int l = strlen (action);
	int count = 0;

	for (int j = 0; j < 256; j++)
	{
		if (!(b = keybindings[j])) continue;

		if (!strncmp (b, action, l))
		{
			twokeys[count++] = j;

			if (count == 2) break;
		}
	}
}


int Menu_KeybindingsCustomDraw (int y)
{
	int keys[2];

	if (keybind_grab)
		M_PrintCenterWhite (y, "Press a Key or Button for this Action");
	else M_PrintCenterWhite (y, "Press ENTER to Change or BACKSPACE to Clear");

	M_PrintCenter (y + 15, DIVIDER_LINE);

	y += 30;

	for (int i = 0; i < NUMBINDINGS; i++)
	{
		// draw the current option highlight wider so that we have room for everything
		if (i == bind_cursor) Menu_HighlightBar (-200, y, 400);

		// action
		M_Print (124 - strlen (key_bindnames[i][1]) * 8, y, key_bindnames[i][1]);

		Menu_FindKeybindingsForAction (key_bindnames[i][0], keys);

		if (keys[0] == -1)
			M_Print (148, y, "(unbound)");
		else
		{
			char *name = Key_KeynumToString (keys[0]);
			M_PrintWhite (148, y, name);

			int x = strlen (name) * 8;

			if (keys[1] != -1)
			{
				M_Print (148 + x + 8, y, "or");
				M_PrintWhite (148 + x + 32, y, Key_KeynumToString (keys[1]));
			}
		}

		// next bind name
		y += 15;
	}

	return y;
}


void Menu_UnbindAction (char *action)
{
	char	*b;

	int l = strlen (action);

	for (int j = 0; j < 256; j++)
	{
		if (!(b = keybindings[j])) continue;
		if (!strncmp (b, action, l)) Key_SetBinding (j, "");
	}
}


// we need to manage the keys for this menu ourselves on account of bindings
void Menu_KeybindingsCustomKey (int key)
{
	if (keybind_grab)
	{
		menu_soundlevel = m_sound_enter;

		// ESC and the console key can never be bound
		if (key == K_ESCAPE)
		{
			// don't change the binding
			keybind_grab = false;
			return;
		}
		else if (key != '`')
		{
			// unbind the key from anything
			Key_SetBinding (key, "");

			// set the new binding
			char cmd[256];
			sprintf (cmd, "bind \"%s\" \"%s\"\n", Key_KeynumToString (key), key_bindnames[bind_cursor][0]);
			Cbuf_InsertText (cmd);
		}

		// exit binding mode
		keybind_grab = false;
		return;
	}

	// get keys for currently selected action
	int keys[2];
	Menu_FindKeybindingsForAction (key_bindnames[bind_cursor][0], keys);

	switch (key)
	{
	case K_ESCAPE:
		menu_Options.EnterMenu ();
		break;

	case K_UPARROW:
		menu_soundlevel = m_sound_nav;
		if ((--bind_cursor) < 0) bind_cursor = NUMBINDINGS - 1;
		break;

	case K_DOWNARROW:
		menu_soundlevel = m_sound_nav;
		if ((++bind_cursor) >= NUMBINDINGS) bind_cursor = 0;
		break;

	case K_ENTER:
		// go into bind mode
		menu_soundlevel = m_sound_query;
		keybind_grab = true;

		// if two keys are bound we delete the first binding
		if (keys[1] != -1) Menu_UnbindAction (key_bindnames[bind_cursor][0]);
		break;

	case K_BACKSPACE:
	case K_DEL:
		menu_soundlevel = m_sound_enter;

		if (keys[0] == -1)
		{
			// nothing to unbind
			menu_soundlevel = m_sound_deny;
		}
		else if (keys[1] == -1)
		{
			// unbind first key
			Key_SetBinding (keys[0], "");
		}
		else
		{
			// unbind second key
			Key_SetBinding (keys[1], "");
		}

		break;
	}
}

void Menu_VideoCustomEnter (void);

void Menu_InitOptionsMenu (void)
{
	// options
	menu_Options.AddOption (new CQMenuBanner ("gfx/p_option.lmp"));
	menu_Options.AddOption (new CQMenuTitle ("Configure Game Options"));
	menu_Options.AddOption (new CQMenuSubMenu ("Customize Controls", &menu_Keybindings));
	menu_Options.AddOption (new CQMenuCommand ("Go to Console", Menu_OptionsGoToConsole));
	menu_Options.AddOption (new CQMenuCommand ("Reset to Defaults", Menu_OptionsResetToDefaults));
	menu_Options.AddOption (new CQMenuCommand ("Save Current Configuration", Host_WriteConfiguration));
	menu_Options.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Options.AddOption (new CQMenuCvarSlider ("Music Volume", &bgmvolume, 0, 1, 0.05));
	menu_Options.AddOption (new CQMenuCvarSlider ("Sound Volume", &volume, 0, 1, 0.05));
	menu_Options.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Options.AddOption (new CQMenuSubMenu ("Video and View Options", &menu_Video));
	menu_Options.AddOption (new CQMenuSubMenu ("Movement and Input Options", &menu_Input));
	menu_Options.AddOption (new CQMenuSubMenu ("Special Effects Options", &menu_Effects));
	menu_Options.AddOption (new CQMenuSubMenu ("Chase Camera Options", &menu_Chase));
	menu_Options.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Options.AddOption (new CQMenuSubMenu ("Content Directories", &menu_ContentDir));

	// chase cam - now that it's fixed, let's expose the wee bugger
	menu_Chase.AddOption (new CQMenuBanner ("gfx/p_option.lmp"));
	menu_Chase.AddOption (new CQMenuTitle ("Chase Camera Options"));
	menu_Chase.AddOption (new CQMenuCvarToggle ("Chase Active", &chase_active, 0, 1));
	menu_Chase.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Chase.AddOption (new CQMenuCvarSlider ("Chase Back", &chase_back, 50, 500, 25));
	menu_Chase.AddOption (new CQMenuCvarSlider ("Chase Up", &chase_up, -64, 256, 16));
	menu_Chase.AddOption (new CQMenuCvarSlider ("Chase Right", &chase_right, -64, 64, 8));
	menu_Chase.AddOption (new CQMenuSpacer (DIVIDER_LINE));
	menu_Chase.AddOption (new CQMenuCvarSlider ("Smoothness", &chase_scale, 0.2, 1.8, 0.1));

	// input
	menu_Input.AddOption (new CQMenuBanner ("gfx/p_option.lmp"));
	menu_Input.AddOption (new CQMenuTitle ("Input Options"));
	menu_Input.AddOption (new CQMenuCvarSlider ("Mouse Speed", &sensitivity, 1, 21, 1));
	menu_Input.AddOption (new CQMenuCvarSlider ("DirectInput Boost", &m_boost, 1, 5, 0.5));
	menu_Input.AddOption (new CQMenuCvarToggle ("Mouse Look", &m_look, 0, 1));
	menu_Input.AddOption (new CQMenuCvarToggle ("Mouse Filter", &m_filter, 0, 1));
	menu_Input.AddOption (new CQMenuCvarToggle ("Invert Mouse", &m_pitch, 0.022, -0.022));
	menu_Input.AddOption (new CQMenuSpacer ());
	menu_Input.AddOption (new CQMenuTitle ("Movement Options"));
	menu_Input.AddOption (new CQMenuCvarToggle ("Always Run", &cl_forwardspeed, 200, 400));
	menu_Input.AddOption (new CQMenuCvarToggle ("Lookspring", &lookspring, 0, 1));
	menu_Input.AddOption (new CQMenuCvarToggle ("Lookstrafe", &lookstrafe, 0, 1));

	// effects
	menu_Effects.AddOption (new CQMenuBanner ("gfx/p_option.lmp"));
	menu_Effects.AddOption (new CQMenuTitle ("Interpolation"));
	menu_Effects.AddOption (new CQMenuCvarToggle ("Orientation", &r_lerporient, 0, 1));
	menu_Effects.AddOption (new CQMenuCvarToggle ("Frame", &r_lerpframe, 0, 1));
	menu_Effects.AddOption (new CQMenuCvarToggle ("Light Style", &r_lerplightstyle, 0, 1));
	menu_Effects.AddOption (new CQMenuSpacer ());
	menu_Effects.AddOption (new CQMenuTitle ("General"));
	menu_Effects.AddOption (new CQMenuCvarToggle ("Flash Blend", &gl_flashblend, 0, 1));

	menu_ContentDir.AddOption (new CQMenuBanner ("gfx/p_option.lmp"));
	menu_ContentDir.AddOption (new CQMenuTitle ("Content Directories"));
	menu_ContentDir.AddOption (new CQMenuCvarTextbox ("Screenshots", &scr_screenshotdir));
	menu_ContentDir.AddOption (new CQMenuCvarTextbox ("Save Games", &host_savedir));
	menu_ContentDir.AddOption (new CQMenuCvarTextbox ("Textures", &gl_texturedir));
	menu_ContentDir.AddOption (new CQMenuCvarTextbox ("MP3 Music", &ds_musicdir));
	menu_ContentDir.AddOption (new CQMenuSpacer ());
	menu_ContentDir.AddOption (new CQMenuTitle ("Warning"));
	menu_ContentDir.AddOption (new CQMenuSpacer ("These aren't as robust as they could be"));
	menu_ContentDir.AddOption (new CQMenuSpacer ("So don't try anything fancy"));

	// keybindings
	menu_Keybindings.AddOption (new CQMenuBanner ("gfx/ttl_cstm.lmp"));
	menu_Keybindings.AddOption (new CQMenuCustomDraw (Menu_KeybindingsCustomDraw));

	// grab every key for these actions
	for (int i = 0; i < 256; i++)
		menu_Keybindings.AddOption (new CQMenuCustomKey (i, Menu_KeybindingsCustomKey));

	// video menu - the rest of the options are deferred until the menu is up!
	// see Menu_VideoCustomEnter in d3d_vidnt.cpp
	menu_Video.AddOption (new CQMenuBanner ("gfx/vidmodes.lmp"));
	menu_Video.AddOption (new CQMenuTitle ("Configure Video Modes"));
	menu_Video.AddOption (new CQMenuCustomEnter (Menu_VideoCustomEnter));
}


/*
========================================================================================================================

					HELP MENU

========================================================================================================================
*/

int menu_HelpPage = 0;
#define	NUM_HELP_PAGES	6


void Menu_HelpCustomEnter (void)
{
	// switch the help page depending on whether we're playing the registered version or not
	if (registered.value)
		menu_HelpPage = 1;
	else menu_HelpPage = 0;
}


int Menu_HelpCustomDraw (int y)
{
	qpic_t *pic = Draw_CachePic (va ("gfx/help%i.lmp", menu_HelpPage));
	Draw_Pic ((vid.width - pic->width) >> 1, y, pic);
	return y + pic->height + 10;
}


void Menu_HelpCustomKey (int key)
{
	menu_soundlevel = m_sound_option;

	switch (key)
	{
	case K_UPARROW:
	case K_RIGHTARROW:
		if (registered.value)
		{
			if (++menu_HelpPage >= NUM_HELP_PAGES)
				menu_HelpPage = 1;
		}
		else
		{
			if (++menu_HelpPage >= NUM_HELP_PAGES)
				menu_HelpPage = 0;
		}
		break;

	case K_DOWNARROW:
	case K_LEFTARROW:
		if (registered.value)
		{
			if (--menu_HelpPage < 1)
				menu_HelpPage = NUM_HELP_PAGES - 1;
		}
		else
		{
			if (--menu_HelpPage < 0)
				menu_HelpPage = NUM_HELP_PAGES - 1;
		}
		break;
	}
}


void Menu_InitHelpMenu (void)
{
	menu_Help.AddOption (new CQMenuCustomEnter (Menu_HelpCustomEnter));
	menu_Help.AddOption (new CQMenuCustomDraw (Menu_HelpCustomDraw));
	menu_Help.AddOption (new CQMenuCustomKey (K_LEFTARROW, Menu_HelpCustomKey));
	menu_Help.AddOption (new CQMenuCustomKey (K_UPARROW, Menu_HelpCustomKey));
	menu_Help.AddOption (new CQMenuCustomKey (K_DOWNARROW, Menu_HelpCustomKey));
	menu_Help.AddOption (new CQMenuCustomKey (K_RIGHTARROW, Menu_HelpCustomKey));
}

