
#include "quakedef.h"
#include "menu_common.h"


// our current menu
CQMenu *menu_Current = NULL;

// sounds to play in menus
menu_soundlevel_t menu_soundlevel = m_sound_none;


char menu_ConBuffer[1024];
float menu_ConBufferTime = 0;
float old_ConBufferTime = 0;

void Menu_PutConsolePrintInbuffer (char *text)
{
	extern cvar_t scr_centertime;

	// copy to console buffer
	strncpy (menu_ConBuffer, text, 1023);

	// display time
	old_ConBufferTime = realtime;

	// give it an extra few seconds so the user has a chance to read it
	menu_ConBufferTime = realtime + 3 + scr_centertime.value;
}


/*
=====================================================================================================================================

				SPIN CONTROL

		Specify NULL for zerotext and units for them to not be displayed if you don't want them.

		If you don't want editing of cvars in-place, make like a textbox and set up a dummy one.

		Two types - one based on a cvar and the other based on a string buffer.  See the sample implementations in the
		multiplayer menu.  The stringbuf should always be NULL terminated.  Look at "SkillNames" in menu_Options.cpp

		Put NULL in command text to center the control options (you're responsible for adding any command text you may
		want yourself...!)

		More logically this should have been 2 separate classes.  Oh well...

=====================================================================================================================================
*/

CQMenuSpinControl::CQMenuSpinControl (char *commandtext, int *menuval, char **stringbuf)
{
	if (commandtext)
		strcpy (this->CommandText, commandtext);
	else this->CommandText[0] = 0;

	this->StringBuf = stringbuf;
	this->MenuVal = menuval;

	// ensure that we're good to begin with
	this->OutputText[0] = 0;

	// value for alternate type
	this->MenuCvar = NULL;

	this->AcceptsInput = true;
}


CQMenuSpinControl::CQMenuSpinControl (char *commandtext, cvar_t *menucvar, float minval, float maxval, float increment, char *zerotext, char *units)
{
	if (commandtext)
		strcpy (this->CommandText, commandtext);
	else this->CommandText[0] = 0;

	this->MenuCvar = menucvar;
	this->MinVal = minval;
	this->MaxVal = maxval;
	this->Increment = increment;

	if (zerotext)
		strcpy (this->ZeroText, zerotext);
	else this->ZeroText[0] = 0;

	if (units)
		strcpy (this->Units, units);
	else this->Units[0] = 0;

	// ensure that we're good to begin with
	this->OutputText[0] = 0;

	// values for alternate type
	this->MenuVal = NULL;
	this->StringBuf = NULL;

	this->AcceptsInput = true;
}


void CQMenuSpinControl::DrawCurrentOptionHighlight (int y)
{
	// this is a temporary hack until we fix a global width for a menu
	Menu_HighlightBar (-175, y, 350);

	bool drawleft = false;
	bool drawright = false;

	if (this->MenuCvar)
	{
		if (this->MenuCvar->value > this->MinVal) drawleft = true;
		if (this->MenuCvar->value < this->MaxVal) drawright = true;
	}
	else if (this->StringBuf)
	{
		if (*(this->MenuVal) > 0) drawleft = true;
		if (this->StringBuf[*(this->MenuVal) + 1]) drawright = true;
	}

	int lpos = 160;

	if (!this->CommandText[0])
	{
		// adjust lpos
		lpos = 148 - strlen (this->OutputText) * 4;
	}

	int rpos = lpos + 16 + strlen (this->OutputText) * 8;

	if (drawleft) M_DrawBackwardsCharacter (lpos, y, 12 + ((int) (realtime * 2) & 1));
	if (drawright) M_DrawCharacter (rpos, y, 12 + ((int) (realtime * 2) & 1));
}


void CQMenuSpinControl::Draw (int y)
{
	// text
	if (this->CommandText[0])
		M_Print (148 - strlen (this->CommandText) * 8, y, this->CommandText);

	if (this->MenuCvar)
	{
		// build the output text
		if (!this->MenuCvar->value && this->ZeroText[0])
			strcpy (this->OutputText, this->ZeroText);
		else
		{
			sprintf (this->OutputText, "%g", this->MenuCvar->value);

			if (this->Units[0] != 0)
			{
				strcat (this->OutputText, " ");
				strcat (this->OutputText, this->Units);
			}
		}
	}
	else if (this->StringBuf)
		strcpy (this->OutputText, this->StringBuf[*(this->MenuVal)]);

	if (!this->CommandText[0])
		M_PrintCenterWhite (y, this->OutputText);
	else M_PrintWhite (172, y, this->OutputText);
}


void CQMenuSpinControl::Key (int k)
{
	if (this->MenuCvar)
	{
		switch (k)
		{
		case K_LEFTARROW:
			menu_soundlevel = m_sound_option;

			this->MenuCvar->value -= this->Increment;

			if (this->MenuCvar->value < this->MinVal)
			{
				this->MenuCvar->value = this->MinVal;
				menu_soundlevel = m_sound_deny;
			}

			break;

		case K_RIGHTARROW:
			menu_soundlevel = m_sound_option;

			this->MenuCvar->value += this->Increment;

			if (this->MenuCvar->value > this->MaxVal)
			{
				this->MenuCvar->value = this->MaxVal;
				menu_soundlevel = m_sound_deny;
			}

			break;
		}
	}
	else if (this->StringBuf)
	{
		switch (k)
		{
		case K_LEFTARROW:
			menu_soundlevel = m_sound_option;

			(*(this->MenuVal))--;

			if (*(this->MenuVal) < 0)
			{
				*(this->MenuVal) = 0;
				menu_soundlevel = m_sound_deny;
			}

			break;

		case K_RIGHTARROW:
			menu_soundlevel = m_sound_option;

			(*(this->MenuVal))++;

			if (!this->StringBuf[*(this->MenuVal)])
			{
				(*(this->MenuVal))--;
				menu_soundlevel = m_sound_deny;
			}

			break;
		}
	}
}


/*
=====================================================================================================================================

				COLOUR BAR

	Because the colour can be a derived value, it takes a pointer rather than a cvar/etc so that it can modify the base value
	directly.  Be certain to either have a base value that you are happy to have modified directly or to set this up correctly!

	(See multiplayer setup menu for an example)

=====================================================================================================================================
*/

CQMenuColourBar::CQMenuColourBar (char *cmdtext, int *colour)
{
	strcpy (this->CommandText, cmdtext);
	this->Colour = colour;
	this->Initial = *(this->Colour);
}


void CQMenuColourBar::PerformEntryFunction (void)
{
	this->Initial = *(this->Colour);
}


bool CQMenuColourBar::IsModified (void)
{
	if (this->Initial == *(this->Colour))
		return false;

	return true;
}


void CQMenuColourBar::Draw (int y)
{
	// text
	M_Print (148 - strlen (this->CommandText) * 8, y, this->CommandText);

	int colour = *(this->Colour);
	int intense = colour * 16 + (colour < 8 ? 11 : 4);

	// colour bar
	for (int i = 0; i < 14; i++)
	{
		// take the approximate midpoint colour (handle backward ranges)
		int c = i * 16 + (i < 8 ? 8 : 7);

		// braw baseline colour (offset downwards a little so that it fits correctly
		Draw_Fill (vid.width / 2 + 12 + i * 8, y + 1, 8, 8, c, 255);
	}

	// draw the highlight rectangle
	Draw_Fill (vid.width / 2 + 11 + colour * 8, y, 10, 10, 15, 255);

	// redraw the highlighted color at brighter intensity (handle backward ranges)
	Draw_Fill (vid.width / 2 + 12 + colour * 8, y + 1, 8, 8, intense, 255);
}


void CQMenuColourBar::Key (int k)
{
	switch (k)
	{
	case K_LEFTARROW:
		menu_soundlevel = m_sound_option;

		(*(this->Colour))--;

		if (*(this->Colour) < 0)
		{
			*(this->Colour) = 0;
			menu_soundlevel = m_sound_deny;
		}

		break;

	case K_RIGHTARROW:
		menu_soundlevel = m_sound_option;

		(*(this->Colour))++;

		if (*(this->Colour) > 13)
		{
			*(this->Colour) = 13;
			menu_soundlevel = m_sound_deny;
		}

		break;
	}
}


/*
=====================================================================================================================================

				CVAR TEXTBOX

		This got real ugly real quick last time I tried to do it.  Let's hope I get it better this time...

		This can either access a string cvar directly (unrecommended) or you can declare but not register a "dummy" version of the
		cvar, create the textbox using that, and set up custom enter and custom key functions to handle changes, or you can do an
		apply menu that also handles the changes... your choice

=====================================================================================================================================
*/

CQMenuCvarTextbox::CQMenuCvarTextbox (char *commandtext, cvar_t *menucvar, bool restrictnumeric)
{
	strncpy (this->CommandText, commandtext, 127);

	this->MenuCvar = menucvar;
	this->AcceptsInput = true;
	this->Modified = false;
	this->InsertMode = true;
	this->IsCurrent = false;
	this->RestrictNumeric = restrictnumeric;
}


bool CQMenuCvarTextbox::IsModified (void)
{
	return this->Modified;
}


void CQMenuCvarTextbox::DrawCurrentOptionHighlight (int y)
{
	// mark as current
	this->IsCurrent = true;

	// highlight bar
	Menu_HighlightBar (-150, y, 300, 14);
}


void CQMenuCvarTextbox::Draw (int y)
{
	// give some space above as well as below
	y += 2;

	// text
	M_Print (148 - strlen (this->CommandText) * 8, y, this->CommandText);

	// textbox
	Draw_Fill (vid.width / 2 - 160 + 168, y - 1, 140, 10, 20, 255);
	Draw_Fill (vid.width / 2 - 160 + 169, y, 138, 8, 0, 192);

	// current text
	for (int i = 0; i < 17; i++)
	{
		if (!this->MenuCvar->string[this->TextStart + i]) break;
		M_DrawCharacter (170 + i * 8, y, this->MenuCvar->string[this->TextStart + i]);
	}

	if (this->IsCurrent)
	{
		// edit cursor
		M_DrawCharacter (170 + 8 * this->TextPos, y, 10 + ((int) (realtime * 4) & 1));

		// insert mode indicator
		M_Print (314, y, this->InsertMode ? "[INS]" : "[OVR]");
	}

	// reset as it may change next frame
	this->IsCurrent = false;
}


void CQMenuCvarTextbox::Key (int k)
{
	// get the real position of the cursor in the string
	int RealTextPos = this->TextPos + this->TextStart;

	switch (k)
	{
	case K_INS:
		// toggle insert mode
		this->InsertMode = !this->InsertMode;
		break;

	case K_LEFTARROW:
		this->TextPos--;
		menu_soundlevel = m_sound_option;

		if (this->TextPos < 0)
		{
			this->TextPos = 0;
			this->TextStart--;
		}

		if (this->TextStart < 0)
		{
			this->TextStart = 0;
			menu_soundlevel = m_sound_deny;
		}
		break;

	case K_RIGHTARROW:
		this->TextPos++;
		menu_soundlevel = m_sound_option;

		if (this->TextPos > strlen (this->MenuCvar->string))
		{
			menu_soundlevel = m_sound_deny;
			this->TextPos = strlen (this->MenuCvar->string);
		}

		if (this->TextPos > 16)
		{
			this->TextPos = 16;
			this->TextStart++;
		}

		if (this->TextStart > strlen (this->MenuCvar->string) - 16)
		{
			this->TextStart = strlen (this->MenuCvar->string) - 16;
			menu_soundlevel = m_sound_deny;
		}
		break;

	case K_HOME:
		// return to start
		menu_soundlevel = m_sound_option;
		this->TextPos = 0;
		this->TextStart = 0;
		break;

	case K_END:
		// go to end
		menu_soundlevel = m_sound_option;
		this->TextPos = 16;
		if (this->TextPos > strlen (this->MenuCvar->string)) this->TextPos = strlen (this->MenuCvar->string);

		this->TextStart = strlen (this->MenuCvar->string) - 16;
		if (this->TextStart < 0) this->TextStart = 0;
		break;

	case K_DEL:
		// prevent deletion if at end of string
		if (RealTextPos >= strlen (this->MenuCvar->string))
		{
			menu_soundlevel = m_sound_deny;
			break;
		}

		// simulate the deletion by moving right then deleting before the cursor
		this->Key (K_RIGHTARROW);
		this->Key (K_BACKSPACE);
		menu_soundlevel = m_sound_option;

		break;

	case K_SHIFT:
		// do nothing
		// this is to prevent holding down shift to type a cap from sounding "deny"
		break;

	case K_BACKSPACE:
		// prevent deletion at start of string
		if (!RealTextPos)
		{
			menu_soundlevel = m_sound_deny;
			break;
		}

		// delete character before cursor
		menu_soundlevel = m_sound_option;
		strcpy (this->ScratchPad, &this->MenuCvar->string[RealTextPos]);
		strcpy (&this->MenuCvar->string[RealTextPos - 1], this->ScratchPad);

		// fix up positioning
		this->TextStart--;

		if (this->TextStart < 0)
		{
			this->TextStart = 0;
			this->TextPos--;
		}

		if (this->TextPos < 0) this->TextPos = 0;
		break;

	default:
		// non alphanumeric
		if (k < 32 || k > 127)
		{
			menu_soundlevel = m_sound_deny;
			break;
		}

		// conservative overflow prevent
		if (strlen (this->MenuCvar->string) > 1020)
		{
			menu_soundlevel = m_sound_deny;
			break;
		}

		if ((k < '0' || k > '9') && this->RestrictNumeric)
		{
			menu_soundlevel = m_sound_deny;
			break;
		}

		menu_soundlevel = m_sound_option;

		if (this->InsertMode)
		{
			// insert mode
			strcpy (this->ScratchPad, &this->MenuCvar->string[RealTextPos]);
			strcpy (&this->MenuCvar->string[RealTextPos + 1], this->ScratchPad);
			this->MenuCvar->string[RealTextPos] = k;

			// move right
			this->Key (K_RIGHTARROW);
		}
		else
		{
			// overwrite mode
			this->MenuCvar->string[RealTextPos] = k;

			// move right
			this->Key (K_RIGHTARROW);
		}

		break;
	}

	// check for modification
	if (stricmp (this->InitialValue, this->MenuCvar->string))
		this->Modified = true;
	else this->Modified = false;
}


void CQMenuCvarTextbox::PerformEntryFunction (void)
{
	// copy cvar string to temp storage
	strcpy (this->InitialValue, this->MenuCvar->string);

	// set positions
	this->TextStart = 0;
	this->TextPos = strlen (this->MenuCvar->string);

	// bound textpos as the string may be longer than the max visible
	// this just runs a K_END on it to get things right
	this->Key (K_END);

	// initially unmodified
	this->Modified = false;
}


int CQMenuCvarTextbox::GetYAdvance (void)
{
	return 20;
}


/*
=====================================================================================================================================

				SUBMENU

=====================================================================================================================================
*/

CQMenuSubMenu::CQMenuSubMenu (char *cmdtext, CQMenu *submenu)
{
	strncpy (this->CommandText, cmdtext, 127);

	this->SubMenu = submenu;
	this->AcceptsInput = true;
}


void CQMenuSubMenu::Draw (int y)
{
	M_PrintCenter (y, this->CommandText);
}


void CQMenuSubMenu::Key (int k)
{
	if (k == K_ENTER) this->SubMenu->EnterMenu ();
}


/*
=====================================================================================================================================

				COMMAND

=====================================================================================================================================
*/

CQMenuCommand::CQMenuCommand (char *cmdtext, menucommand_t command)
{
	strncpy (this->CommandText, cmdtext, 127);

	if (command)
		this->Command = command;
	else this->Command = Menu_NullCommand;

	this->AcceptsInput = true;
}


void CQMenuCommand::Draw (int y)
{
	M_PrintCenter (y, this->CommandText);
}


void CQMenuCommand::Key (int k)
{
	if (k == K_ENTER)
	{
		menu_soundlevel = m_sound_option;
		this->Command ();
	}
}


/*
=====================================================================================================================================

				APPLY OPTION

=====================================================================================================================================
*/

CQMenuApplyOption::CQMenuApplyOption (char *cmdtext, menucommand_t command)
{
	strncpy (this->CommandText, cmdtext, 127);

	if (command)
		this->Command = command;
	else this->Command = Menu_NullCommand;

	// this won't accept input until apply is set on
	this->AcceptsInput = false;
}


void CQMenuApplyOption::SetApplyOn (void)
{
	this->AcceptsInput = true;
}


void CQMenuApplyOption::SetApplyOff (void)
{
	this->AcceptsInput = false;
}


void CQMenuApplyOption::DrawCurrentOptionHighlight (int y)
{
	Menu_HighlightBar (y + 15);
}


void CQMenuApplyOption::Draw (int y)
{
	// don't draw unless it's switched on
	if (!this->AcceptsInput) return;

	M_PrintCenter (y, DIVIDER_LINE);
	M_PrintCenter (y + 15, this->CommandText);
}


int CQMenuApplyOption::GetYAdvance (void)
{
	return 30;
}


void CQMenuApplyOption::Key (int k)
{
	if (k == K_ENTER)
	{
		menu_soundlevel = m_sound_option;
		this->Command ();
	}
}


/*
=====================================================================================================================================

				SPACER

		Either a blank space or some text.

		Set to NULL for space.

=====================================================================================================================================
*/

CQMenuSpacer::CQMenuSpacer (char *spacertext)
{
	if (spacertext)
		strcpy (this->CommandText, spacertext);
	else this->CommandText[0] = 0;

	this->AcceptsInput = false;
}


void CQMenuSpacer::Draw (int y)
{
	if (this->CommandText[0])
	{
		M_PrintCenter (y, this->CommandText);
	}
}


int CQMenuSpacer::GetYAdvance ()
{
	if (this->CommandText[0])
		return 15;

	return 5;
}


/*
=====================================================================================================================================

				CUSTOM ENTER

		Allows menus to define their own enter functions outside of the common framework

=====================================================================================================================================
*/

CQMenuCustomEnter::CQMenuCustomEnter (menucommand_t enterfunc)
{
	this->EnterFunc = enterfunc;
	this->AcceptsInput = false;
}


void CQMenuCustomEnter::PerformEntryFunction (void)
{
	this->EnterFunc ();
}


int CQMenuCustomEnter::GetYAdvance (void)
{
	return 0;
}


/*
=====================================================================================================================================

				CUSTOM DRAW

		Allows menus to define their own rendering outside of the common framework

		Note: the custom draw func should return the amount by which to advance Y rather than Y itself, a bit confusing
		so I'll probably change it...

=====================================================================================================================================
*/

CQMenuCustomDraw::CQMenuCustomDraw (imenucommandi_t customdrawfunc)
{
	this->CustomDrawFunc = customdrawfunc;
	this->AcceptsInput = false;
}


void CQMenuCustomDraw::Draw (int y)
{
	this->Y = this->CustomDrawFunc (y) - y;
}


int CQMenuCustomDraw::GetYAdvance (void)
{
	return this->Y;
}


/*
=====================================================================================================================================

				CUSTOM KEY

		Allows menus to define their own key capture events outside of the common framework

=====================================================================================================================================
*/

CQMenuCustomKey::CQMenuCustomKey (int keycapture, menucommand_t capturefunc, bool fwoverride)
{
	// single key handler function
	this->KeyCapture = keycapture;
	this->CaptureFuncNoKeyArg = capturefunc;
	this->CaptureFuncKeyArg = NULL;
	this->AcceptsInput = false;

	// allow to override or combine with the standard key func
	this->OverRideFramework = fwoverride;
}


CQMenuCustomKey::CQMenuCustomKey (int keycapture, menucommandi_t capturefunc, bool fwoverride)
{
	// shared key handler function
	this->KeyCapture = keycapture;
	this->CaptureFuncKeyArg = capturefunc;
	this->CaptureFuncNoKeyArg = NULL;
	this->AcceptsInput = false;

	// allow to override or combine with the standard key func
	this->OverRideFramework = fwoverride;
}


bool CQMenuCustomKey::CheckStatus (void *stuff)
{
	// retrieve the key
	int k = *((int *) stuff);

	// execute the capture function
	if (k == this->KeyCapture)
	{
		// execute the custom key func and notify that caller that we got it
		// the key arg version allows a shared handler
		if (this->CaptureFuncKeyArg)
			this->CaptureFuncKeyArg (k);
		else this->CaptureFuncNoKeyArg ();

		// return
		return this->OverRideFramework;
	}

	// didn't get it
	return false;
}


int CQMenuCustomKey::GetYAdvance ()
{
	return 0;
}


/*
=====================================================================================================================================

				TITLE

=====================================================================================================================================
*/

CQMenuTitle::CQMenuTitle (char *title)
{
	strncpy (this->CommandText, title, 127);
	this->AcceptsInput = false;
}


void CQMenuTitle::Draw (int y)
{
	M_PrintCenterWhite (y, this->CommandText);
	M_PrintCenter (y + 15, DIVIDER_LINE);
}


int CQMenuTitle::GetYAdvance (void)
{
	return 30;
}


/*
=====================================================================================================================================

				BANNER

=====================================================================================================================================
*/

CQMenuBanner::CQMenuBanner (char *pic)
{
	this->Pic = (char *) malloc (strlen (pic) + 1);
	strcpy (this->Pic, pic);
	this->AcceptsInput = false;
}


void CQMenuBanner::Draw (int y)
{
	// we can't load this on init as the filesystem may not be up yet
	qpic_t *pic = Draw_CachePic (this->Pic);

	// draw centered on screen and offset down
	Draw_Pic ((vid.width - pic->width) >> 1, y, pic);
	this->Y = pic->height + 10;
}


int CQMenuBanner::GetYAdvance (void)
{
	return this->Y;
}


/*
=====================================================================================================================================

				CVAR SLIDER

=====================================================================================================================================
*/

CQMenuCvarSlider::CQMenuCvarSlider (char *commandtext, cvar_t *menucvar, float minval, float maxval, float stepsize)
{
	strncpy (this->CommandText, commandtext, 127);

	this->AcceptsInput = true;

	this->MenuCvar = menucvar;
	this->SliderMin = minval;
	this->SliderMax = maxval;
	this->StepSize = stepsize;

	// check for inverse range
	if (this->SliderMin > this->SliderMax) this->StepSize *= -1;
}


void CQMenuCvarSlider::Draw (int y)
{
	// text
	M_Print (148 - strlen (this->CommandText) * 8, y, this->CommandText);

	// slider left
	M_DrawCharacter (168, y + 1, 128);

	// slider body
	for (int i = 0; i < 10; i++)
		M_DrawCharacter (176 + i * 8, y + 1, 129);

	// slider right
	M_DrawCharacter (256, y + 1, 130);

	// evaluate point on the range
	float point = (this->MenuCvar->value - this->SliderMin) / (this->SliderMax - this->SliderMin);

	// don't go beyond the bounds
	if (point < 0) point = 0;
	if (point > 1) point = 1;

	// draw indicator
	M_DrawCharacter (172 + point * 80, y + 1, 131);
}


void CQMenuCvarSlider::Key (int k)
{
	// store value out
	float val = this->MenuCvar->value;

	switch (k)
	{
	case K_LEFTARROW:
		val -= this->StepSize;
		menu_soundlevel = m_sound_option;

		if (val < this->SliderMin && this->SliderMin < this->SliderMax)
		{
			// normal range
			menu_soundlevel = m_sound_deny;
			val = this->SliderMin;
		}
		else if (val > this->SliderMin && this->SliderMin > this->SliderMax)
		{
			// normal range
			menu_soundlevel = m_sound_deny;
			val = this->SliderMin;
		}

		Cvar_Set (this->MenuCvar, val);
		break;

	case K_ENTER:
	case K_RIGHTARROW:
		val += this->StepSize;
		menu_soundlevel = m_sound_option;

		if (val > this->SliderMax && this->SliderMin < this->SliderMax)
		{
			// normal range
			menu_soundlevel = m_sound_deny;
			val = this->SliderMax;
		}
		else if (val < this->SliderMax && this->SliderMin > this->SliderMax)
		{
			// inverse range
			menu_soundlevel = m_sound_deny;
			val = this->SliderMax;
		}

		Cvar_Set (this->MenuCvar, val);
		break;

	default:
		break;
	}
}


/*
=====================================================================================================================================

				CVAR ON/OFF TOGGLE

=====================================================================================================================================
*/

CQMenuCvarToggle::CQMenuCvarToggle (char *commandtext, cvar_t *menucvar, float toggleoffvalue, float toggleonvalue)
{
	strncpy (this->CommandText, commandtext, 127);

	this->AcceptsInput = true;
	this->MenuCvar = menucvar;
	this->ToggleOffValue = toggleoffvalue;
	this->ToggleOnValue = toggleonvalue;

	// set togglestate depending on which of on/off the value is closest to
	this->ToggleState = !((toggleonvalue - menucvar->value) > (menucvar->value - toggleoffvalue));
}


void CQMenuCvarToggle::DrawCurrentOptionHighlight (int y)
{
	Menu_HighlightBar (y);

	// this is a hack to get the l/r indicators working correctly here
	Menu_DrawOption (172, y, this->ToggleState ? "  " : "   ", true, true);
}


void CQMenuCvarToggle::Draw (int y)
{
	M_Print (148 - strlen (this->CommandText) * 8, y, this->CommandText);

	// update togglestate otherwise it will only reflect the initial cvar value
	// set togglestate depending on which of on/off the value is closest to
	this->ToggleState = !((this->ToggleOnValue - this->MenuCvar->value) > (this->MenuCvar->value - this->ToggleOffValue));

	// this is a hack to get the l/r indicators working correctly here
	Menu_DrawOption (172, y, this->ToggleState ? "On" : "Off", false, false);
}


void CQMenuCvarToggle::Key (int k)
{
	switch (k)
	{
	case K_LEFTARROW:
	case K_RIGHTARROW:
	case K_ENTER:
		// flip the toggle state
		this->ToggleState = !this->ToggleState;
		menu_soundlevel = m_sound_option;

		// set the cvar
		Cvar_Set (this->MenuCvar, this->ToggleState ? this->ToggleOnValue : this->ToggleOffValue);
		break;

	default:
		break;
	}
}


/*
=====================================================================================================================================

				EMPTY OPTION

=====================================================================================================================================
*/

// these are called if none of the virtual methods are overloaded
// standard y advance
int CQMenuOption::GetYAdvance (void) {return 15;}

// draw nothing
void CQMenuOption::Draw (int y) {}

// no key events
void CQMenuOption::Key (int k) {}

// whether it can accept input or not
bool CQMenuOption::CanAcceptInput (void) {return this->AcceptsInput;}

// used for checking status of something
bool CQMenuOption::CheckStatus (void *stuff) {return false;}

// option number - never overridden
void CQMenuOption::SetOptionNumber (int optionnum) {this->OptionNum = optionnum;}
int CQMenuOption::GetOptionNumber (void) {return this->OptionNum;}

// this one lets each type of option perform a custom function when the menu is entered
// it might be it's own internal validation of cvar constraints or something entirely different
void CQMenuOption::PerformEntryFunction (void) {}

// this one lets options call public members from their parent menu
void CQMenuOption::SetParentMenu (CQMenu *parent) {this->Parent = parent;}

// this one lets a derived class report if it's value has been modified
bool CQMenuOption::IsModified (void) {return false;}

// this one lets options position their own highlight bars in cases where they draw anything nonstandard
void CQMenuOption::DrawCurrentOptionHighlight (int y) {Menu_HighlightBar (y);}

/*
=====================================================================================================================================

				THE MENU ITSELF

=====================================================================================================================================
*/

CQMenu::CQMenu (CQMenu *previous, m_state_t menustate)
{
	// defaults
	this->PreviousMenu = previous;
	this->MenuOptions = NULL;
	this->CurrentOption = NULL;
	this->NumOptions = 0;
	this->MenuState = menustate;
	this->ApplyOption = NULL;
	this->NeedsApply = false;
}


void CQMenu::AddApplyCommand (char *commandtext, menucommand_t applyfunc)
{
	// you can add as many of these as you like, but only the first one added will be effective
	if (this->ApplyOption) return;

	// hmm - decision time... add this as an option which we can selectively disable or keep it separate and get a mess in the key func...
	// ok, decision made it appears... still created a bit of a mess in the classes themselves though, but done is done...
	this->ApplyOption = new CQMenuApplyOption (commandtext, applyfunc);
	this->AddOption (this->ApplyOption);
}


void CQMenu::AddOption (CQMenuOption *opt)
{
	// next points to NULL always so that we can identify the end of the list
	opt->NextOption = NULL;

	// set option number and increment options counter
	opt->SetOptionNumber (this->NumOptions++);

	// set parent menu
	opt->SetParentMenu (this);

	// the first option that can accept input is the initial current option
	if (!this->CurrentOption && opt->CanAcceptInput ())
		this->CurrentOption = opt;

	if (!this->MenuOptions)
	{
		// set at first in the list
		this->MenuOptions = opt;

		// previous points to itself as there is only one item in the list
		opt->PrevOption = this->MenuOptions;
		return;
	}

	// find the end of the list
	for (CQMenuOption *opt2 = this->MenuOptions; opt2; opt2 = opt2->NextOption)
	{
		if (!opt2->NextOption)
		{
			// add to the end
			opt2->NextOption = opt;
			opt->PrevOption = opt2;

			// update first item previous pointer
			this->MenuOptions->PrevOption = opt;
			break;
		}
	}
}


void CQMenu::Draw (void)
{
	int y = 25;
	bool Modified = false;

	// no options
	if (!this->MenuOptions) return;

	// check for NULL current option and set it to the first option
	if (!this->CurrentOption) this->CurrentOption = this->MenuOptions;

	// ensure that current option can accept input
	if (!this->CurrentOption->CanAcceptInput ())
	{
		int scrollcount = 0;

		// move current option to one that can accept input, unless we can't find one, in which case it's
		// likely custom handled...
		while (1)
		{
			// go to the next option
			this->CurrentOption = this->CurrentOption->NextOption;

			// check for null and reset to start of list if so
			if (!this->CurrentOption) this->CurrentOption = this->MenuOptions;

			// break if it accepts input
			if (this->CurrentOption->CanAcceptInput ()) break;

			// check for a menu with no options that accept input
			if (++scrollcount > this->NumOptions) break;
		}
	}

	// draw all of the options
	for (CQMenuOption *opt = this->MenuOptions; opt; opt = opt->NextOption)
	{
		// check for a modified option
		if (opt->IsModified ()) Modified = true;

		if (opt->CanAcceptInput () && (opt->GetOptionNumber () == this->CurrentOption->GetOptionNumber ()))
		{
			// draw the highlight bar under it
			opt->DrawCurrentOptionHighlight (y);

			// the fact that it's current needs special handling in come draw funcs
			opt->Draw (y);
		}
		else opt->Draw (y);

		// get the next y position
		y += opt->GetYAdvance ();
	}

	// note - this won't kick in until the following frame
	if (Modified && this->ApplyOption)
		this->ApplyOption->SetApplyOn ();
	else if (!Modified && this->ApplyOption)
		this->ApplyOption->SetApplyOff ();

	// draw the console buffer
	if (menu_ConBufferTime > realtime)
	{
		menu_ConBufferTime -= (realtime - old_ConBufferTime);
		old_ConBufferTime = realtime;

		M_PrintCenter (vid.height - 32, menu_ConBuffer);
	}
}


void CQMenu::Key (int k)
{
	if (!this->MenuOptions)
	{
		switch (k)
		{
		case K_ESCAPE:
			// to do - back out to previous menu or quit menus if no previous defined
			if (this->PreviousMenu)
				this->PreviousMenu->EnterMenu ();
			else
			{
				// exit the menus
			}
			break;

		default:
			break;
		}

		return;
	}

	// look for any custom key functions that match this key and override the framework if so
	for (CQMenuOption *opt = this->MenuOptions; opt; opt = opt->NextOption)
		if (opt->CheckStatus (&k))
			return;

	// depending on how fast (or clumsy!) we are with the keyboard, this function can be triggered before
	// the draw function, so we need to ensure that we get a valid current option in here too...
	if (!this->CurrentOption) this->CurrentOption = this->MenuOptions;

	// count of menu scrolling
	int scrollcount = 0;

	// send key events through the framework
	switch (k)
	{
	case K_ESCAPE:
		// to do - back out to previous menu or quit menus if no previous defined
		if (this->PreviousMenu)
			this->PreviousMenu->EnterMenu ();
		else
		{
			// exit the menus
		}
		break;

	case K_UPARROW:
		while (1)
		{
			// go to the previous option
			this->CurrentOption = this->CurrentOption->PrevOption;

			// break if it accepts input
			if (this->CurrentOption->CanAcceptInput ()) break;

			// check for a menu with no options that accept input
			if (++scrollcount > this->NumOptions) break;
		}

		menu_soundlevel = m_sound_nav;
		break;

	case K_DOWNARROW:
		while (1)
		{
			// go to the next option
			this->CurrentOption = this->CurrentOption->NextOption;

			// check for null and reset to start of list if so
			if (!this->CurrentOption) this->CurrentOption = this->MenuOptions;

			// break if it accepts input
			if (this->CurrentOption->CanAcceptInput ()) break;

			// check for a menu with no options that accept input
			if (++scrollcount > this->NumOptions) break;
		}

		menu_soundlevel = m_sound_nav;
		break;

	default:
		// send it through the option key handler
		if (this->CurrentOption->CanAcceptInput ())
			this->CurrentOption->Key (k);

		break;
	}
}


void CQMenu::EnterMenu (void)
{
	key_dest = key_menu;
	m_state = this->MenuState;
	menu_soundlevel = m_sound_enter;

	// perform all entry functions for all options
	// these are left till after the common state as they may override elements of it
	for (CQMenuOption *opt = this->MenuOptions; opt; opt = opt->NextOption)
		opt->PerformEntryFunction ();

	// store to current pointer for easier access
	menu_Current = this;
}


/*
=====================================================================================================================================

				SCROLLBOX PROVIDER

=====================================================================================================================================
*/


CScrollBoxProvider::CScrollBoxProvider (int numitems, int maxvisibleitems, int maxitemwidth)
{
	this->NumItems = numitems;
	this->MaxVisibleItems = maxvisibleitems;
	this->ScrollBoxStartItem = 0;
	this->ScrollBoxCurrentItem = 0;

	// set event callbacks to nothing yet
	this->DrawItemCallback = NULL;
	this->HoverItemCallback = NULL;
	this->EnterItemCallback = NULL;
	this->MaxItemWidth = maxitemwidth;
}


void CScrollBoxProvider::SetDrawItemCallback (menucommandii_t drawitemcallback)
{
	this->DrawItemCallback = drawitemcallback;
}


void CScrollBoxProvider::SetHoverItemCallback (menucommandiii_t hoveritemcallback)
{
	this->HoverItemCallback = hoveritemcallback;
}


void CScrollBoxProvider::SetEnterItemCallback (menucommandi_t enteritemcallback)
{
	this->EnterItemCallback = enteritemcallback;
}


int CScrollBoxProvider::DrawItems (int x, int starty)
{
	if (!this->DrawItemCallback) return starty;

	int initialy = starty;

	// draw the textbox
	Draw_TextBox (x, starty, this->MaxItemWidth * 8 + 8, this->MaxVisibleItems * 12 + 4);
	Draw_VScrollBar (x + this->MaxItemWidth * 8 + 8, starty + 8, this->MaxVisibleItems * 12 + 4, this->ScrollBoxCurrentItem, this->NumItems);

	starty += 12;

	for (int i = this->ScrollBoxStartItem; i < this->NumItems; i++)
	{
		// check for end of list
		if (i - this->ScrollBoxStartItem >= this->MaxVisibleItems)
			break;

		// check for active item and run the hover callback if any
		if (i == this->ScrollBoxCurrentItem)
			if (this->HoverItemCallback)
				this->HoverItemCallback (initialy, starty, i);

		// draw the current item
		this->DrawItemCallback (starty, i);
		starty += 12;
	}

	// testing...
	/*
	M_Print (-50, vid.height - 60, va ("ScrollBoxStartItem:   %i\n", this->ScrollBoxStartItem));
	M_Print (-50, vid.height - 50, va ("ScrollBoxCurrentItem: %i\n", this->ScrollBoxCurrentItem));
	M_Print (-50, vid.height - 40, va ("MaxVisibleItems:      %i\n", this->MaxVisibleItems));
	M_Print (-50, vid.height - 30, va ("NumItems:             %i\n", this->NumItems));
	*/

	return starty;
}


void CScrollBoxProvider::KeyFunc (int key)
{
	switch (key)
	{
	case K_ENTER:
		if (this->EnterItemCallback)
			this->EnterItemCallback (this->ScrollBoxCurrentItem);
		break;

	case K_UPARROW:
		this->ScrollBoxCurrentItem--;

		// simple scrollbox
		if (this->NumItems <= this->MaxVisibleItems)
		{
			if (this->ScrollBoxCurrentItem < 0)
				this->ScrollBoxCurrentItem = this->NumItems - 1;

			break;
		}

		// check for wrap
		if (this->ScrollBoxCurrentItem < 0)
		{
			this->ScrollBoxStartItem = (this->NumItems - this->MaxVisibleItems);
			this->ScrollBoxCurrentItem = this->NumItems - 1;
		}

		// pull back the start item - note - current item can never go to -1 (trapped above)
		// so start item will never go below 0
		if (this->ScrollBoxCurrentItem < this->ScrollBoxStartItem)
			this->ScrollBoxStartItem--;

		break;

	case K_DOWNARROW:
		this->ScrollBoxCurrentItem++;

		// simple scrollbox
		if (this->NumItems <= this->MaxVisibleItems)
		{
			if (this->ScrollBoxCurrentItem >= this->NumItems)
				this->ScrollBoxCurrentItem = 0;

			break;
		}

		// check for advance
		if (this->ScrollBoxCurrentItem >= this->MaxVisibleItems)
		{
			// advance the start item
			this->ScrollBoxStartItem++;
		}

		// don't let start advance too far
		if (this->ScrollBoxStartItem > this->NumItems - this->MaxVisibleItems)
			this->ScrollBoxStartItem = this->NumItems - this->MaxVisibleItems;

		if (this->ScrollBoxCurrentItem >= this->NumItems)
		{
			// reset to start
			this->ScrollBoxCurrentItem = 0;
			this->ScrollBoxStartItem = 0;
		}

		break;
	}
}


/*
=====================================================================================================================================

				SUPPORT ROUTINES

=====================================================================================================================================
*/


void Menu_DrawOption (int x, int y, char *option, bool leftflash, bool rightflash)
{
	// do the option
	M_PrintWhite (x, y, option);

	if (leftflash) M_DrawBackwardsCharacter (x - 12, y, 12 + ((int) (realtime * 2) & 1));
	if (rightflash) M_DrawCharacter (x + 4 + strlen (option) * 8, y, 12 + ((int) (realtime * 2) & 1));
}


void Menu_DrawOption (int x, int y, char *option)
{
	// no flashes
	Menu_DrawOption (x, y, option, false, false);
}


void Menu_HighlightBar (int xofs, int y, int width, int height)
{
	int colorramp[] = {64, 65, 66, 67, 68, 67, 66, 65};

	int color = colorramp[(int) (realtime * 10) % 8];

	Draw_Fill ((vid.width / 2) + xofs, y - 1, width, height, color, 192);
}


void Menu_HighlightBar (int xofs, int y, int width)
{
	Menu_HighlightBar (xofs, y, width, 12);
}


void Menu_HighlightBar (int y, int width)
{
	Menu_HighlightBar (-150, y, width, 12);
}


void Menu_HighlightBar (int y)
{
	Menu_HighlightBar (-150, y, 300, 12);
}


// this is just provided as a development aid to have something to stick into unfinished options
void Menu_NullCommand (void) { /* does nothing */ }


bool Menu_MessageBox (char *text)
{
	// temp - we'll do this right later on!
	return SCR_ModalMessage (text);

	if (cls.state == ca_dedicated)
		return true;

	return false;
}


void Menu_InitHelpMenu (void);
void Menu_InitOptionsMenu (void);
void Menu_InitSPMenu (void);
void Menu_InitMainMenu (void);
void Menu_InitSaveLoadMenu (void);
void Menu_InitMultiplayerMenu (void);

void Menu_CommonInit (void)
{
	// create the menus.
	// menus *MUST* be declared as non-pointers, otherwise you *MUST* be *VERY* *CAREFUL* about the order
	// in which you declare/instantiate/add options so that there are valid submenu/previous menu pointers
	// all the time.  just declare them as non-pointers, OK.
	Menu_InitMainMenu ();
	Menu_InitSPMenu ();
	Menu_InitSaveLoadMenu ();
	Menu_InitMultiplayerMenu ();
	Menu_InitOptionsMenu ();
	Menu_InitHelpMenu ();
}


