
typedef enum
{
	m_sound_none,
	m_sound_enter,
	m_sound_nav,
	m_sound_option,
	m_sound_query,
	m_sound_deny
} menu_soundlevel_t;

extern menu_soundlevel_t menu_soundlevel;

// these will go away when we finally remove the last of the old menu code.
// for now only save/load and some of the multiplayer ones are needed.
// i've added m_other so that i can define new options any time without
// having to add new items to this list...
typedef enum m_state_t
{
	m_none,
	m_main,
	m_load,
	m_save,
	m_multiplayer,
	m_setup,
	m_net,
	m_options,
	m_video,
	m_keys,
	m_help,
	m_quit,
	m_serialconfig,
	m_modemconfig,
	m_lanconfig_newgame,
	m_lanconfig_joingame,
	m_gameoptions,
	m_search,
	m_slist,
	m_other
} m_state_t;

extern m_state_t m_state;

typedef void (*menucommand_t) (void);
typedef int (*imenucommand_t) (void);
typedef int (*imenucommandi_t) (int);
typedef void (*menucommandii_t) (int, int);
typedef void (*menucommandiii_t) (int, int, int);
typedef void (*menucommandi_t) (int);

void Menu_CommonInit (void);
void Menu_NullCommand (void);


class CQMenuOption
{
public:
	virtual void Draw (int y);
	virtual int GetYAdvance (void);
	virtual void Key (int k);
	virtual bool CheckStatus (void *stuff);
	virtual void PerformEntryFunction (void);
	virtual bool IsModified (void);
	virtual void DrawCurrentOptionHighlight (int y);
	bool CanAcceptInput (void);
	CQMenuOption *NextOption;
	CQMenuOption *PrevOption;

	int GetOptionNumber (void);
	void SetOptionNumber (int optionnum);
	void SetParentMenu (class CQMenu *parent);

protected:
	char CommandText[128];
	bool AcceptsInput;

private:
	class CQMenu *Parent;
	int OptionNum;
};


class CQMenuCvar
{
protected:
	cvar_t *MenuCvar;
	bool Modified;
};


class CQMenuColourBar : public CQMenuOption
{
public:
	CQMenuColourBar (char *cmdtext, int *colour);
	void Draw (int y);
	void Key (int k);
	bool IsModified (void);
	void PerformEntryFunction (void);

private:
	int *Colour;
	int Initial;
};


class CQMenuApplyOption : public CQMenuOption
{
public:
	CQMenuApplyOption (char *cmdtext, menucommand_t command);
	void Draw (int y);
	void Key (int k);
	void SetApplyOn (void);
	void SetApplyOff (void);
	int GetYAdvance (void);
	void DrawCurrentOptionHighlight (int y);

private:
	menucommand_t Command;
};


class CQMenuSpinControl : public CQMenuOption, public CQMenuCvar
{
public:
	CQMenuSpinControl (char *commandtext, cvar_t *menucvar, float minval, float maxval, float increment, char *zerotext, char *units);
	CQMenuSpinControl (char *commandtext, int *menuval, char **stringbuf);
	void Draw (int y);
	void Key (int k);
	void DrawCurrentOptionHighlight (int y);

private:
	float MinVal;
	float MaxVal;
	float Increment;
	int *MenuVal;
	char **StringBuf;
	char ZeroText[64];
	char Units[64];
	char OutputText[256];
};


class CQMenuCvarToggle : public CQMenuOption, public CQMenuCvar
{
public:
	CQMenuCvarToggle (char *commandtext, cvar_t *menucvar, float toggleoffvalue = 0, float toggleonvalue = 1);
	void Draw (int y);
	void Key (int k);
	void DrawCurrentOptionHighlight (int y);

private:
	bool ToggleState;
	float ToggleOnValue;
	float ToggleOffValue;
};


class CQMenuCvarSlider : public CQMenuOption, public CQMenuCvar
{
public:
	CQMenuCvarSlider (char *commandtext, cvar_t *menucvar, float minval, float maxval, float stepsize);
	void Draw (int y);
	void Key (int k);

private:
	float SliderMin;
	float SliderMax;
	float StepSize;
};


class CQMenuCvarTextbox : public CQMenuOption, public CQMenuCvar
{
public:
	CQMenuCvarTextbox (char *commandtext, cvar_t *menucvar, bool restrictnumeric = false);
	void Draw (int y);
	void Key (int k);
	void PerformEntryFunction (void);
	int GetYAdvance (void);
	bool IsModified (void);
	void DrawCurrentOptionHighlight (int y);

private:
	char InitialValue[1024];
	char ScratchPad[1024];

	// properties
	int TextStart;
	int TextPos;

	// insert or overwrite
	bool InsertMode;

	// true if the current option
	bool IsCurrent;

	// if true input can only be numeric
	bool RestrictNumeric;
};


class CQMenuCustomEnter : public CQMenuOption
{
public:
	CQMenuCustomEnter (menucommand_t enterfunc);
	void PerformEntryFunction (void);
	int GetYAdvance (void);

private:
	menucommand_t EnterFunc;
};


class CQMenuSpacer : public CQMenuOption
{
public:
	CQMenuSpacer (char *spacertext = NULL);
	void Draw (int y);
	int GetYAdvance (void);
};


class CQMenuCustomDraw : public CQMenuOption
{
public:
	CQMenuCustomDraw (imenucommandi_t customdrawfunc);
	void Draw (int y);
	int GetYAdvance (void);

private:
	imenucommandi_t CustomDrawFunc;
	int Y;
};


class CQMenuCustomKey : public CQMenuOption
{
public:
	CQMenuCustomKey (int keycapture, menucommand_t capturefunc, bool fwoverride = true);
	CQMenuCustomKey (int keycapture, menucommandi_t capturefunc, bool fwoverride = true);
	bool CheckStatus (void *stuff);
	int GetYAdvance (void);

private:
	int KeyCapture;
	bool OverRideFramework;
	menucommand_t CaptureFuncNoKeyArg;
	menucommandi_t CaptureFuncKeyArg;
};


class CQMenuBanner : public CQMenuOption
{
public:
	CQMenuBanner (char *pic);
	void Draw (int y);
	int GetYAdvance (void);

private:
	char *Pic;
	int Y;
};


class CQMenuTitle : public CQMenuOption
{
public:
	CQMenuTitle (char *title);
	void Draw (int y);
	int GetYAdvance (void);
};


class CQMenuCommand : public CQMenuOption
{
public:
	CQMenuCommand (char *cmdtext, menucommand_t command);
	void Draw (int y);
	void Key (int k);

private:
	menucommand_t Command;
};


class CQMenu
{
public:
	CQMenu (CQMenu *previous, m_state_t menustate);
	void Draw (void);
	void Key (int k);
	void AddOption (CQMenuOption *opt);
	void AddApplyCommand (char *commandtext, menucommand_t applyfunc);
	void EnterMenu (void);
	CQMenuOption *MenuOptions;

private:
	char ApplyText[128];
	CQMenuApplyOption *ApplyOption;
	bool NeedsApply;
	int NumOptions;
	CQMenuOption *CurrentOption;
	CQMenu *PreviousMenu;
	m_state_t MenuState;
};


class CQMenuSubMenu : public CQMenuOption
{
public:
	CQMenuSubMenu (char *cmdtext, CQMenu *submenu);
	void Draw (int y);
	void Key (int k);

private:
	CQMenu *SubMenu;
};


class CScrollBoxProvider
{
public:
	CScrollBoxProvider (int numitems, int maxvisibleitems, int maxitemwidth);
	void SetDrawItemCallback (menucommandii_t drawitemcallback);
	void SetHoverItemCallback (menucommandiii_t hoveritemcallback);
	void SetEnterItemCallback (menucommandi_t enteritemcallback);
	void KeyFunc (int key);
	int DrawItems (int x, int starty);

private:
	// callbacks - 2 args; which item number it happened on and the vertical position it happened on
	menucommandii_t DrawItemCallback;
	menucommandiii_t HoverItemCallback;
	menucommandi_t EnterItemCallback;

	// properties
	int ScrollBoxStartItem;
	int ScrollBoxCurrentItem;
	int NumItems;
	int MaxVisibleItems;
	int MaxItemWidth;
};


// support routines
bool Menu_MessageBox (char *text);


// menus
extern CQMenu menu_Main;
extern CQMenu menu_Singleplayer;
extern CQMenu menu_Save;
extern CQMenu menu_Load;
extern CQMenu menu_Multiplayer;
extern CQMenu menu_TCPIPNewGame;
extern CQMenu menu_TCPIPJoinGame;
extern CQMenu menu_GameConfig;
extern CQMenu menu_Setup;
extern CQMenu menu_Options;
extern CQMenu menu_Keybindings;
extern CQMenu menu_Video;
extern CQMenu menu_Help;

// our current menu
extern CQMenu *menu_Current;


// drawing functions - relative to 320 width
void M_Print (int cx, int cy, char *str);
void M_PrintWhite (int cx, int cy, char *str);
void M_PrintCenter (int cy, char *str);
void M_PrintCenterWhite (int cy, char *str);
void Menu_HighlightBar (int y);
void Menu_HighlightBar (int y, int width);
void Menu_HighlightBar (int xofs, int y, int width);
void Menu_HighlightBar (int xofs, int y, int width, int height);
void Menu_DrawOption (int x, int y, char *option);
void Menu_DrawOption (int x, int y, char *option, bool leftflash, bool rightflash);
void M_DrawCharacter (int cx, int line, int num);
void M_DrawBackwardsCharacter (int cx, int line, int num);


